/* $Id: sofi_img_domeflat.c,v 1.17 2013-03-12 08:04:52 llundin Exp $
 *
 * This file is part of the SOFI Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:04:52 $
 * $Revision: 1.17 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <cpl.h>

#include "irplib_utils.h"

#include "sofi_utils.h"
#include "sofi_pfits.h"
#include "sofi_dfs.h"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int sofi_img_domeflat_create(cpl_plugin *);
static int sofi_img_domeflat_exec(cpl_plugin *);
static int sofi_img_domeflat_destroy(cpl_plugin *);
static int sofi_img_domeflat(cpl_parameterlist *, cpl_frameset *);
static cpl_image * sofi_img_domeflat_compute(cpl_frameset *);
static cpl_image * sofi_img_domeflat_bias(cpl_image *, cpl_image *,
        int, int, int, int);
static int sofi_img_domeflat_save(cpl_image *, cpl_parameterlist *, 
        cpl_frameset *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char sofi_img_domeflat_description[] = 
"sofi_img_domeflat -- SOFI imaging flat-field creation.\n"
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"raw-file.fits "SOFI_IMG_DOMEFLAT_RAW"\n";

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe));
    cpl_plugin  *   plugin = &recipe->interface;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    SOFI_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "sofi_img_domeflat",
                    "Dome flat recipe",
                    sofi_img_domeflat_description,
                    "Yves Jung",
                    "yjung@eso.org",
                    sofi_get_license(),
                    sofi_img_domeflat_create,
                    sofi_img_domeflat_exec,
                    sofi_img_domeflat_destroy);

    cpl_pluginlist_append(list, plugin);
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
static int sofi_img_domeflat_create(cpl_plugin * plugin)
{
    cpl_recipe      * recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();

    /* Fill the parameters list */
    
    /* Return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int sofi_img_domeflat_exec(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    return sofi_img_domeflat(recipe->parameters, recipe->frames);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int sofi_img_domeflat_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    cpl_parameterlist_delete(recipe->parameters);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int sofi_img_domeflat(
        cpl_parameterlist   *   parlist, 
        cpl_frameset        *   framelist)
{
    cpl_frameset    *   flatframes;
    cpl_image       *   flat;
    
    /* Identify the RAW and CALIB frames in the input frameset */
    if (sofi_dfs_set_groups(framelist)) {
        cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames");
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Retrieve raw frames */
    if ((flatframes = sofi_extract_frameset(framelist,
                    SOFI_IMG_DOMEFLAT_RAW)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot find flat frames in the input list");
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Compute the flat field */
    cpl_msg_info(cpl_func, "Compute dome flat");
    if ((flat = sofi_img_domeflat_compute(flatframes)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot compute dome flat");
        cpl_frameset_delete(flatframes);
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_frameset_delete(flatframes);
    
    /* Save the flat field */
    cpl_msg_info(cpl_func, "Save the product");
    if (sofi_img_domeflat_save(flat, parlist, framelist) == -1) {
        cpl_msg_error(cpl_func, "Cannot save the product");
        cpl_image_delete(flat);
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_image_delete(flat);

    /* Free and return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the flatfield
  @param    flatframes  the input frames with 8 frames
  @return   the flat field or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_image * sofi_img_domeflat_compute(cpl_frameset * fset)
{
    cpl_imagelist       *   ilist;
    int                     x1, x2, x3, x4;
    cpl_image           *   on;
    cpl_image           *   on_ma;
    cpl_image           *   off;
    cpl_image           *   off_ma;
    cpl_image           *   bias;
    
    /* Test entries */
    if (fset == NULL) return NULL;
    if (cpl_frameset_get_size(fset) != 8) return NULL;
    
    /* Initialise */
    x1 = 500;
    x2 = 600;
    x3 = 50;
    x4 = 150;
    
    /* Load the input frames */
    cpl_msg_info(cpl_func, "Load the frames");
    if ((ilist = cpl_imagelist_load_frameset(fset, CPL_TYPE_FLOAT, 0,
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot load the frames");
        return NULL;
    }

    /* Get the on images */
    on =    cpl_image_add_create(   cpl_imagelist_get(ilist, 3),
                                    cpl_imagelist_get(ilist, 4));
    cpl_image_divide_scalar(on, 2);
    on_ma = cpl_image_add_create(   cpl_imagelist_get(ilist, 2),
                                    cpl_imagelist_get(ilist, 5));
    cpl_image_divide_scalar(on_ma, 2);
    
    /* Get the on bias */
    if ((bias = sofi_img_domeflat_bias(on, on_ma, x1, x2, x3, x4)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot compute the bias of the on frames");
        cpl_image_delete(on);
        cpl_image_delete(on_ma);
        cpl_imagelist_delete(ilist);
        return NULL;
    }
    cpl_image_delete(on_ma);

    /* Correct the on bias */
    cpl_image_subtract(on, bias);
    cpl_image_delete(bias);

    /* Get the off images */
    off =   cpl_image_add_create(   cpl_imagelist_get(ilist, 0),
                                    cpl_imagelist_get(ilist, 7));
    cpl_image_divide_scalar(off, 2);
    off_ma =cpl_image_add_create(   cpl_imagelist_get(ilist, 1),
                                    cpl_imagelist_get(ilist, 6));
    cpl_image_divide_scalar(off_ma, 2);
    cpl_imagelist_delete(ilist);
 
    /* Get the off bias */
    if ((bias = sofi_img_domeflat_bias(off, off_ma, x1, x2, x3, 
                    x4)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot compute the bias of the off frames");
        cpl_image_delete(on);
        cpl_image_delete(off);
        cpl_image_delete(off_ma);
        return NULL;
    }
    cpl_image_delete(off_ma);

    /* Correct the on bias */
    cpl_image_subtract(off, bias);
    cpl_image_delete(bias);

    /* Compute the flat */
    cpl_image_subtract(on, off);
    cpl_image_delete(off);

    /* Normalise the flat */
    cpl_image_normalise(on, CPL_NORM_MEAN);
    
    /* Return */
    return on;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the bias
  @param    in      the input image
  @param    in_mask the input image
  @param    x1
  @param    x2
  @param    x3
  @param    x4
  @return   The image with the bias removed or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_image * sofi_img_domeflat_bias(
        cpl_image           *   in,
        cpl_image           *   in_mask,
        int                     x1,
        int                     x2,
        int                     x3,
        int                     x4)
{
    int                     nx, ny;
    cpl_image           *   a_1d;
    cpl_image           *   b_1d;
    cpl_image           *   c_1d;
    cpl_image           *   a_2d;
    float               *   pa_1d;
    float               *   pa_2d;
    int                     i, j;

    /* Test entries */
    if (in == NULL) return NULL;
    if (in_mask == NULL) return NULL;

    /* Initialise */
    nx = cpl_image_get_size_x(in);
    ny = cpl_image_get_size_y(in);

    /* Compute the bias */
    a_1d = cpl_image_collapse_window_create(in, x1, 1, x2, ny, 1);
    cpl_image_divide_scalar(a_1d, x2-x1+1);
    c_1d = cpl_image_collapse_window_create(in_mask, x1, 1, x2, ny, 1);
    cpl_image_divide_scalar(c_1d, x2-x1+1);
    b_1d = cpl_image_collapse_window_create(in_mask, x3, 1, x4, ny, 1);
    cpl_image_divide_scalar(b_1d, x4-x3+1);
    cpl_image_subtract(a_1d, c_1d);
    cpl_image_delete(c_1d);
    cpl_image_add(a_1d, b_1d);
    cpl_image_delete(b_1d);
    
    /* Expand a_1d to a_2d */
    a_2d = cpl_image_new(nx, ny, CPL_TYPE_FLOAT);
    pa_2d = cpl_image_get_data_float(a_2d);
    pa_1d = cpl_image_get_data_float(a_1d);
    for (j=0 ; j<ny ; j++) {
        for (i=0 ; i<nx ; i++) {
            pa_2d[i+j*nx] = pa_1d[j];
        }
    }
    cpl_image_delete(a_1d);

    return a_2d;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the lampflat recipe products on disk
  @param    flat        the flat image produced
  @param    parlist     the input list of parameters
  @param    set         the input frame set
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int sofi_img_domeflat_save(
        cpl_image           *   flat,
        cpl_parameterlist   *   parlist,
        cpl_frameset        *   set)
{
    cpl_propertylist    *   plist;
    cpl_propertylist    *   paflist;
    cpl_propertylist    *   qclist;
    const cpl_frame     *   ref_frame;


    /* Get the QC params in qclist */
    qclist = cpl_propertylist_new();

    /* Write the average image */
    irplib_dfs_save_image(set,
            parlist,
            set,
            flat,
            CPL_BPP_IEEE_FLOAT,
            "sofi_img_domeflat",
            SOFI_IMG_DOMEFLAT_RES,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            "sofi_img_domeflat.fits");
    
    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);

    /* Get FITS header from reference file */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist,
        "^(DATE-OBS|ESO DET CHIP NAME|ARCFILE|ESO TPL ID|ESO DET MODE NAME|"
        "ESO DET NCORRS NAME|ESO DET RSPEED|ESO DET DIT)$", 0);
    cpl_propertylist_delete(plist);

    /* Copy the QC in paflist */
    /* cpl_propertylist_copy_property_regexp(paflist, qclist, ".", 0) ; */
    cpl_propertylist_delete(qclist);

    /* Save the PAF file */
    cpl_dfs_save_paf("SOFI",
            "sofi_img_domeflat",
            paflist,
            "sofi_img_domeflat.paf");
    cpl_propertylist_delete(paflist);
    return  0;
}

