/* $Id: sofi_img_illum.c,v 1.24 2013-03-12 08:04:52 llundin Exp $
 *
 * This file is part of the SOFI Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:04:52 $
 * $Revision: 1.24 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <cpl.h>

#include "irplib_plugin.h"
#include "irplib_utils.h"
#include "irplib_calib.h"
#include "irplib_strehl.h"

#include "sofi_utils.h"
#include "sofi_pfits.h"
#include "sofi_dfs.h"

/*-----------------------------------------------------------------------------
                            Defines
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "sofi_img_illum"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static cpl_image * sofi_img_illum_reduce(cpl_frameset *, const char *,
                                         const char *, const char *,
                                         cpl_table **);
static cpl_bivector * sofi_img_illum_find_pos(const cpl_imagelist *,
                                              const cpl_bivector *);
static cpl_vector * sofi_img_illum_phot(cpl_imagelist *, cpl_bivector *,int *);
static
cpl_error_code sofi_img_illum_save(cpl_frameset *, const cpl_image *,
                                   const cpl_table *, const cpl_parameterlist *);

cpl_recipe_define(sofi_img_illum, SOFI_BINARY_VERSION,
                  "Lars Lundin", PACKAGE_BUGREPORT, "2002,2003,2009", 
                  "SOFI imaging illumination frame recipe",
                  RECIPE_STRING " -- SOFI imaging illumination frame recipe.\n"
                  "The files listed in the Set Of Frames (sof-file) " 
                  "must be tagged:\n"
                  "raw-file.fits "SOFI_IMG_ILLUM_RAW" or\n"
                  "flat-file.fits "SOFI_CALIB_FLAT" or\n"
                  "dark-file.fits "SOFI_CALIB_DARK"\n");

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int         s_hx;
    int         s_hy;
    double      star_r;
    double      bg_r1;
    double      bg_r2;
    int         pos_x;
    int         pos_y;

    /* Outputs */
    cpl_size    poly_deg;
    double      illum1;
    double      illumx;
    double      illumy;
    double      illumxy;
    double      illumxx;
    double      illumyy;
} sofi_img_illum_config;


/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Fill the recipe parameterlist
  @param    self  The parameterlist
  @return   0 iff everything is ok

 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code sofi_img_illum_fill_parameterlist(cpl_parameterlist * self)
{
    const char * context = PACKAGE "." RECIPE_STRING;
    cpl_error_code err;

    cpl_ensure_code(self, CPL_ERROR_NULL_INPUT);

    /* Fill the parameters list */

    /* --star_r */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING, "star_r",
                                          10.0, NULL, context,
                                          "The star radius [pixel]");
    cpl_ensure_code(!err, err);

    /* --bg_r1 */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING, "bg_r1",
                                          12.0, NULL, context, "The internal "
                                          "background radius [pixel]");
    cpl_ensure_code(!err, err);

    /* --bg_r2 */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING, "bg_r2",
                                          30.0, NULL, context, "The external "
                                          "background radius [pixel]");
    cpl_ensure_code(!err, err);

    /* --s_hx */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING, "s_hx",
                                       50, NULL, context, "Half-size of the "
                                       "search box in X [pixel]");
    cpl_ensure_code(!err, err);

    /* --s_hy */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING, "s_hy",
                                       50, NULL, context, "Half-size of the "
                                       "search box in Y [pixel]");
    cpl_ensure_code(!err, err);

    /* --pos_x */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING, "pos_x",
                                       -1, NULL, context, "The star "
                                       "X-position [pixel]");
    cpl_ensure_code(!err, err);

    /* --pos_y */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING, "pos_y",
                                       -1, NULL, context, "The star "
                                       "Y-position [pixel]");
    cpl_ensure_code(!err, err);

    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    The recipe data reduction part is implemented here 
  @param    framelist   the frames list
  @param    parlist     the parameters list
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static int sofi_img_illum(cpl_frameset            * framelist,
                          const cpl_parameterlist * parlist)
{
    cpl_frameset * rawframes = NULL;
    cpl_image    * ima = NULL;
    cpl_table    * flux = NULL;
    const char   * flat;
    const char   * dark;


    /* Retrieve input parameters */
    sofi_img_illum_config.s_hx
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "s_hx");
    sofi_img_illum_config.s_hy
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "s_hy");
    sofi_img_illum_config.star_r
        = irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING,
                                          "star_r");
    sofi_img_illum_config.bg_r1
        = irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING,
                                          "bg_r1");
    sofi_img_illum_config.bg_r2
        = irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING,
                                          "bg_r2");
    sofi_img_illum_config.pos_x
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "pos_x");
    sofi_img_illum_config.pos_y
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "pos_y");

    bug_if(0);

    /* Identify the RAW and CALIB frames in the input frameset */
    skip_if (sofi_dfs_set_groups(framelist));

    /* Retrieve calibration data */
    flat = sofi_extract_filename(framelist, SOFI_CALIB_FLAT);
    dark = sofi_extract_filename(framelist, SOFI_CALIB_DARK);

    /* Retrieve raw frames */
    rawframes = sofi_extract_frameset(framelist, SOFI_IMG_ILLUM_RAW);
    error_if (rawframes == NULL, CPL_ERROR_DATA_NOT_FOUND,
              "Cannot find any raw frame in the input");

    /* Apply the data reduction */
    cpl_msg_info(cpl_func, "Apply the reduction");
    cpl_msg_indent_more();
    ima = sofi_img_illum_reduce(rawframes, flat, dark, NULL, &flux);
    cpl_msg_indent_less();

    skip_if (ima == NULL);
  
    /* Save the products */
    cpl_msg_info(cpl_func, "Saving the products");
    skip_if (sofi_img_illum_save(framelist, ima, flux, parlist));

    end_skip;

    cpl_image_delete(ima);
    cpl_table_delete(flux);
    cpl_frameset_delete(rawframes);

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    The recipe data reduction part is implemented here 
  @param    raw         the raw frames
  @param    flat        the flat field or NULL
  @param    dark        the dark or NULL
  @param    bpm         the bad pixels map or NULL
  @param    fl          the returned flux table
  @return   the illumination image or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_image * sofi_img_illum_reduce(
        cpl_frameset    *   raw,
        const char      *   flat,
        const char      *   dark,
        const char      *   bpm,
        cpl_table       **  fl)
{
    cpl_bivector    *   offsets;
    cpl_imagelist   *   in;
    cpl_bivector    *   positions;
    cpl_matrix      *   purged_pos;
    double          *   positions_x,
                    *   positions_y;
    cpl_vector      *   flux;
    cpl_vector      *   purged_flux;
    cpl_polynomial  *   poly;
    cpl_image       *   poly_ima;
    int                 nval, nx, ny;
    cpl_size            power[2];
    int                 i, j;
    
    /* Initialise */
    *fl = NULL;

    /* Get the offsets from the raw frames */
    cpl_msg_info(cpl_func, "Get the offsets from the files headers");
    cpl_msg_indent_more();
    if ((offsets = sofi_get_offsets(raw)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot get the offsets");
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_msg_indent_less();

    /* Load the input data */
    cpl_msg_info(cpl_func, "Load the input data");
    cpl_msg_indent_more();
    if ((in = cpl_imagelist_load_frameset(raw, CPL_TYPE_FLOAT, 1, 0)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot load input data");
        cpl_bivector_delete(offsets);
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_msg_indent_less();

    /* Apply the calibrations */
    if (flat || dark || bpm) {
        cpl_msg_info(cpl_func, "Apply the calibrations");
        cpl_msg_indent_more();
        if (irplib_flat_dark_bpm_calib(in, flat, dark, bpm) == -1) {
            cpl_msg_error(cpl_func, "Cannot calibrate the data");
            cpl_imagelist_delete(in);
            cpl_bivector_delete(offsets);
            cpl_msg_indent_less();
            return NULL;
        }
        cpl_msg_indent_less();
    }

    /* Find the positions */
    cpl_msg_info(cpl_func, "Find the positions");
    cpl_msg_indent_more();
    if ((positions=sofi_img_illum_find_pos(in, offsets)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot find the positions");
        cpl_imagelist_delete(in);
        cpl_bivector_delete(offsets);
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_bivector_delete(offsets);
    cpl_msg_indent_less();

    /* Compute the photometry */
    cpl_msg_info(cpl_func, "Compute the photometry");
    cpl_msg_indent_more();
    if ((flux=sofi_img_illum_phot(in, positions, &nval)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot find the positions");
        cpl_imagelist_delete(in);
        cpl_bivector_delete(positions);
        cpl_msg_indent_less();
        return NULL;
    }
    nx = cpl_image_get_size_x(cpl_imagelist_get(in, 0));
    ny = cpl_image_get_size_y(cpl_imagelist_get(in, 0));
    cpl_imagelist_delete(in);
    cpl_msg_indent_less();
    if (nval < 1) {
        cpl_msg_error(cpl_func, "No flux computed");
        cpl_bivector_delete(positions);
        cpl_vector_delete(flux);
        return NULL;
    }
    
    /* Purge positions */
    purged_pos = cpl_matrix_new(2, nval);
    positions_x = cpl_bivector_get_x_data(positions);
    positions_y = cpl_bivector_get_y_data(positions);
    purged_flux = cpl_vector_new(nval);
    j = 0;
    for (i=0; i<cpl_vector_get_size(flux); i++) {
        if (fabs(cpl_vector_get(flux, i)) > 0) {
            cpl_matrix_set(purged_pos, 0, j, positions_x[i]);
            cpl_matrix_set(purged_pos, 1, j, positions_y[i]);
            cpl_vector_set(purged_flux, j, cpl_vector_get(flux, i));
            j++;
        }
    }

    /* Create the flux table */
    *fl = cpl_table_new(cpl_vector_get_size(flux));
    cpl_table_new_column(*fl, "POSX", CPL_TYPE_DOUBLE);
    cpl_table_new_column(*fl, "POSY", CPL_TYPE_DOUBLE);
    cpl_table_new_column(*fl, "FLUX", CPL_TYPE_DOUBLE);
    for (i=0; i<cpl_vector_get_size(flux); i++) {
        cpl_table_set_double(*fl, "POSX", i, positions_x[i]);
        cpl_table_set_double(*fl, "POSY", i, positions_y[i]);
        cpl_table_set_double(*fl, "FLUX", i, cpl_vector_get(flux, i));
    }
    cpl_bivector_delete(positions);
    cpl_vector_delete(flux);
    
    /* Degree of the 2D-polynomial */
    if (nval >= 6) {
        sofi_img_illum_config.poly_deg = 2;
    } else if (nval >= 3) {
        sofi_img_illum_config.poly_deg = 1;
    } else {
        sofi_img_illum_config.poly_deg = 0;
    }
    cpl_msg_info(cpl_func, "Fitting a %d-degree 2D-polynomial to %d points",
                 (int)sofi_img_illum_config.poly_deg, nval);
    poly = cpl_polynomial_new(2);
    if (cpl_polynomial_fit(poly, purged_pos, NULL, purged_flux, NULL,
                           CPL_FALSE, NULL, &(sofi_img_illum_config.poly_deg))) {
        cpl_matrix_delete(purged_pos);
        cpl_vector_delete(purged_flux);
        cpl_error_set_where(cpl_func);
        cpl_polynomial_delete(poly);
        cpl_table_delete(*fl);
        *fl = NULL;
        return NULL;
    }
    cpl_matrix_delete(purged_pos);
    cpl_vector_delete(purged_flux);
   
    /* Fill the coeffs */
    power[0] = 0; power[1] = 0;
    sofi_img_illum_config.illum1 = cpl_polynomial_get_coeff(poly, power);
    if (sofi_img_illum_config.poly_deg >= 1) {
        power[0] = 1; power[1] = 0;
        sofi_img_illum_config.illumx = cpl_polynomial_get_coeff(poly, power);
        power[0] = 0; power[1] = 1;
        sofi_img_illum_config.illumy = cpl_polynomial_get_coeff(poly, power);
    }
    if (sofi_img_illum_config.poly_deg >= 2) {
        power[0] = 1; power[1] = 1;
        sofi_img_illum_config.illumxy = cpl_polynomial_get_coeff(poly, power);
        power[0] = 2; power[1] = 0;
        sofi_img_illum_config.illumxx = cpl_polynomial_get_coeff(poly, power);
        power[0] = 0; power[1] = 2;
        sofi_img_illum_config.illumyy = cpl_polynomial_get_coeff(poly, power);
    }
    cpl_msg_info(cpl_func, "P(X,Y)=a0+a1*X+a2*Y+a3*X*Y+a4*X^2+a5*Y^2");
    cpl_msg_info(cpl_func, "a0 = %g", sofi_img_illum_config.illum1);
    cpl_msg_info(cpl_func, "a1 = %g", sofi_img_illum_config.illumx);
    cpl_msg_info(cpl_func, "a2 = %g", sofi_img_illum_config.illumy);
    cpl_msg_info(cpl_func, "a3 = %g", sofi_img_illum_config.illumxy);
    cpl_msg_info(cpl_func, "a4 = %g", sofi_img_illum_config.illumxx);
    cpl_msg_info(cpl_func, "a5 = %g", sofi_img_illum_config.illumyy);
    
    /* Create the polynomial image */
    poly_ima = cpl_image_new(nx, ny, CPL_TYPE_FLOAT);
    cpl_image_fill_polynomial(poly_ima, poly, 1.0, 1.0, 1.0, 1.0);
    cpl_polynomial_delete(poly);

    /* Normalise the image */
    cpl_image_normalise(poly_ima, CPL_NORM_MEAN);

    return poly_ima;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Compute the photometry of the bright star in the images.
  @param    ilist       the images list
  @param    positions   the star positions
  @param    nvalid      number of valid points
  @return   the flux in a vector or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * sofi_img_illum_phot(
        cpl_imagelist   *   ilist,
        cpl_bivector    *   positions,
        int             *   nvalid)
{
    cpl_vector          *   flux;
    cpl_image           *   cur_ima;
    double              *   positions_x;
    double              *   positions_y;
    int                     ni;
    int                     nok;
    double                  bg, fl;
    int                     i;

    /* Test entries */
    if (ilist == NULL || positions == NULL || nvalid == NULL) return NULL;
    ni = cpl_imagelist_get_size(ilist);
    positions_x = cpl_bivector_get_x_data(positions);
    positions_y = cpl_bivector_get_y_data(positions);
    nok = 0;

    /* Create the flux vector */
    flux = cpl_vector_new(ni);

    /* Loop on the frames */
    for (i=0; i<ni; i++) {
        cur_ima = cpl_imagelist_get(ilist, i);
        bg = irplib_strehl_ring_background(cur_ima, (int)positions_x[i],
                (int)positions_y[i], sofi_img_illum_config.bg_r1,
                sofi_img_illum_config.bg_r2, IRPLIB_BG_METHOD_MEDIAN);
        fl = irplib_strehl_disk_flux(cur_ima, (int)positions_x[i],
                (int)positions_y[i], sofi_img_illum_config.star_r, bg);
        cpl_vector_set(flux, i, fl);
        if (fabs(fl)>1e-3) nok++;
        cpl_msg_info(cpl_func, "Flux for image %d: %g", i+1, fl);
    }
    *nvalid = nok;
    return flux;
}
 
/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Get the positions of the bright star in the images.
  @param    ilist       the images list
  @param    offsets     the offsets between frames
  @return   the positions or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_bivector * sofi_img_illum_find_pos(const cpl_imagelist * ilist,
                                              const cpl_bivector  * offsets)
{
    const int      ni = cpl_imagelist_get_size(ilist);
    const int      nx = cpl_image_get_size_x(cpl_imagelist_get_const(ilist, 0));
    const int      ny = cpl_image_get_size_y(cpl_imagelist_get_const(ilist, 0));
    cpl_bivector * positions;
    double       * positions_x;
    double       * positions_y;
    const double * offsets_x = cpl_bivector_get_x_data_const(offsets);
    const double * offsets_y = cpl_bivector_get_y_data_const(offsets);
    cpl_image    * img_window = NULL;
    cpl_image    * img_filter = NULL;
    cpl_mask     * kernel = NULL;
    int            i;


    cpl_ensure(ilist   != NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(offsets != NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(cpl_bivector_get_size(offsets) == ni,
               CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
    
    /* Create the positions bivector */
    positions = cpl_bivector_new(ni);
    positions_x = cpl_bivector_get_x_data(positions);
    positions_y = cpl_bivector_get_y_data(positions);

    /* Create the kernel */
    kernel = cpl_mask_new(3, 3);
    cpl_mask_not(kernel);
    
    /* Loop on all the images */
    for (i=0; i < ni; i++) {
        const cpl_image * image = cpl_imagelist_get_const(ilist, i);

        /* Define zone */
        const int pos_x = sofi_img_illum_config.pos_x > 0
            ? sofi_img_illum_config.pos_x : nx/2;
        const int pos_y = sofi_img_illum_config.pos_y > 0
            ? sofi_img_illum_config.pos_y : ny/2;

        const int llx = pos_x + offsets_x[i] - sofi_img_illum_config.s_hx;
        const int urx = pos_x + offsets_x[i] + sofi_img_illum_config.s_hx;
        const int lly = pos_y + offsets_y[i] - sofi_img_illum_config.s_hy;
        const int ury = pos_y + offsets_y[i] + sofi_img_illum_config.s_hy;

        cpl_size        posx, posy;

        /* Test zone */
        if (llx > urx || lly > ury || llx < 1 || urx > nx || lly < 1 ||
            ury > ny) {
            (void)cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                                        "Bad specified zone: llx=%d, urx=%d, "
                                        "lly=%d, ury=%d", llx, urx, lly, ury);
            break;
        }


        /* Extract */
        cpl_image_delete(img_window);
        cpl_image_delete(img_filter);
        img_window = cpl_image_extract(image, llx, lly, urx, ury);
        img_filter = cpl_image_new(1 + urx - llx, 1 + ury - lly,
                                   cpl_image_get_type(img_window));

        cpl_image_filter_mask(img_filter, img_window, kernel,
                              CPL_FILTER_MEDIAN, CPL_BORDER_FILTER);

        /* Find the max */
        cpl_image_get_maxpos(img_filter, &posx, &posy);
        positions_x[i] = llx + posx;
        positions_y[i] = lly + posy;
        cpl_msg_info(cpl_func, "Star found at pos %g, %g in image %d",
                     positions_x[i], positions_y[i], i+1);

    }
    cpl_image_delete(img_window);
    cpl_image_delete(img_filter);
    cpl_mask_delete(kernel);
    if (i != ni) {
        cpl_bivector_delete(positions);
        positions = NULL;
    }

    return positions;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Save the different products of the recipe
  @param    set         the input frame set
  @param    ima         the fitted image
  @param    flux        the flux table (optional)
  @param    parlist     the input list of parameters
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code sofi_img_illum_save(cpl_frameset            * set,
                                          const cpl_image         * ima,
                                          const cpl_table         * flux,
                                          const cpl_parameterlist * parlist)
{
    const cpl_frame     *   ref_frame;
    cpl_propertylist    *   plist;
    cpl_propertylist    *   paflist;
    cpl_propertylist    *   qclist;

    /* Get the QC params in qclist */
    qclist = cpl_propertylist_new();
    cpl_propertylist_append_double(qclist, "ESO QC ILLUM1",
            sofi_img_illum_config.illum1);
    cpl_propertylist_append_double(qclist, "ESO QC ILLUMX", 
            sofi_img_illum_config.illumx);
    cpl_propertylist_append_double(qclist, "ESO QC ILLUMY", 
            sofi_img_illum_config.illumy);
    cpl_propertylist_append_double(qclist, "ESO QC ILLUMXY",
            sofi_img_illum_config.illumxy);
    cpl_propertylist_append_double(qclist, "ESO QC ILLUMXX", 
            sofi_img_illum_config.illumxx);
    cpl_propertylist_append_double(qclist, "ESO QC ILLUMYY",
            sofi_img_illum_config.illumyy);

    /* Write the image */
    irplib_dfs_save_image(set,
            parlist,
            set,
            ima,
            CPL_BPP_IEEE_FLOAT,
            "sofi_img_illum",
            SOFI_IMG_ILLUM_RES,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            "sofi_img_illum.fits");

    /* Write the table */
    if (flux != NULL) {
        irplib_dfs_save_table(set,
                parlist,
                set,
                flux,
                NULL,
                "sofi_img_illum",
                SOFI_IMG_ILLUM_TAB,
                qclist,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                "sofi_img_illum_flux.fits");
    }

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);

    /* Get FITS header from reference file */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return cpl_error_set_where(cpl_func);
    }

    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist,
        "^(ARCFILE|MJD-OBS|ESO TPL ID|DATE-OBS|ESO DET DIT|ESO DET NDIT|"
        "ESO DET NCORRS|ESO DET MODE NAME)$", 0);
    cpl_propertylist_delete(plist);

    /* Copy the QC in paflist */
    cpl_propertylist_copy_property_regexp(paflist, qclist, ".", 0);
    cpl_propertylist_delete(qclist);

    /* Save the PAF file */
    cpl_dfs_save_paf("SOFI",
            "sofi_img_illum",
            paflist,
            "sofi_img_illum.paf");
    cpl_propertylist_delete(paflist);
    return  cpl_error_get_code();
}
