/* $Id: sofi_img_jitter.c,v 1.30 2013-03-12 08:04:51 llundin Exp $
 *
 * This file is part of the SOFI Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:04:51 $
 * $Revision: 1.30 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <string.h>
#include <cpl.h>

#include "irplib_utils.h"
#include "irplib_calib.h"

#include "sofi_utils.h"
#include "sofi_pfits.h"
#include "sofi_dfs.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define NEGLIG_OFF_DIFF     0.1
#define SQR(x) ((x)*(x))

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int sofi_img_jitter_create(cpl_plugin *);
static int sofi_img_jitter_exec(cpl_plugin *);
static int sofi_img_jitter_destroy(cpl_plugin *);
static int sofi_img_jitter(cpl_parameterlist *, cpl_frameset *);

static cpl_image ** sofi_img_jitter_reduce(cpl_frameset *, cpl_frameset *, 
        const char *, const char *, const char *, cpl_vector **);
static cpl_imagelist ** sofi_img_jitter_load(cpl_frameset *, cpl_frameset *);
static cpl_vector * sofi_img_jitter_sky(cpl_imagelist **, cpl_imagelist *);
static cpl_vector * sofi_img_jitter_sky_running(cpl_imagelist **);
static cpl_image ** sofi_img_jitter_saa(cpl_imagelist *, 
        cpl_frameset *);
static int sofi_img_jitter_sub_row_median(cpl_image *);
static cpl_table * sofi_img_jitter_qc(cpl_image *);
static double sofi_img_jitter_get_mode(cpl_vector *);
static int sofi_img_jitter_save(cpl_image *, cpl_table *, cpl_vector *, 
        cpl_parameterlist *, cpl_frameset *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    const char      *   offsets;
    const char      *   objects;
    int                 crosstalk;
    int                 sky_minnb;
    int                 sky_halfw;
    int                 sky_rejmin;
    int                 sky_rejmax;
    int                 sx;
    int                 sy;
    int                 mx;
    int                 my;
    cpl_geom_combine    comb_meth;
    int                 rej_low;
    int                 rej_high;
    int                 row_med;
    /* Outputs */
    double          pixscale;
    double          dit;
    int             nb_obj_frames;
    int             nb_sky_frames;
    int             nb_rej_frames;
    double          iq;
    int             nbobjs;
    double          fwhm_pix;
    double          fwhm_arcsec;
    double          fwhm_mode;
} sofi_img_jitter_config;

static char sofi_img_jitter_description[] =
"sofi_img_jitter -- SOFI imaging jitter recipe.\n"
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"raw-file.fits "SOFI_IMG_JITTER_OBJ_RAW" or\n"
"raw-file.fits "SOFI_IMG_JITTER_SKY_RAW" or\n"
"flat-file.fits "SOFI_CALIB_FLAT" or\n"
"dark-file.fits "SOFI_CALIB_DARK"\n";

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe));
    cpl_plugin  *   plugin = &recipe->interface;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    SOFI_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "sofi_img_jitter",
                    "Jitter recipe",
                    sofi_img_jitter_description,
                    "Yves Jung",
                    "yjung@eso.org",
                    sofi_get_license(),
                    sofi_img_jitter_create,
                    sofi_img_jitter_exec,
                    sofi_img_jitter_destroy);

    cpl_pluginlist_append(list, plugin);
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
static int sofi_img_jitter_create(cpl_plugin * plugin)
{
    cpl_recipe      * recipe;
    cpl_parameter   * p;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();

    /* Fill the parameters list */
    /* --off */
    p = cpl_parameter_new_value("sofi.sofi_img_jitter.offsets", CPL_TYPE_STRING,
                                "Offsets file. If the i'th frame is believed "
                                "to have the offset (CumoffsetX_i, "
                                "CumoffsetY_i), then the i'th line in "
                                "the offset file should be:\n "
                                "X_i Y_i\n "
                                "where\n "
                                "X_i = CumoffsetX_0 - CumoffsetX_i\n "
                                "and\n "
                                "Y_i = CumoffsetY_0 - CumoffsetY_i.\n "
                                "(This implies that the first line in the "
                                "offset file should consist of two zeroes).",
                                "sofi.sofi_img_jitter", NULL);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "off");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);

    /* --objs */
    p = cpl_parameter_new_value("sofi.sofi_img_jitter.objects", 
            CPL_TYPE_STRING, "objects file", "sofi.sofi_img_jitter", NULL);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "objs");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);

    /* --crosstalk */
    p = cpl_parameter_new_value("sofi.sofi_img_jitter.crosstalk", CPL_TYPE_BOOL,
            "flag to remove the crosstalk effect", "sofi.sofi_img_jitter",
            TRUE);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "crosstalk");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    
    /* --sky_par */
    p = cpl_parameter_new_value("sofi.sofi_img_jitter.sky_par",
                                CPL_TYPE_STRING,
                                "Rejection parameters for sky filtering",
                                "sofi.sofi_img_jitter",
                                "10,7,3,3");
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "sky_par");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);

    /* --xcorr */
    p = cpl_parameter_new_value("sofi.sofi_img_jitter.xcorr",
                                CPL_TYPE_STRING,
                                "Cross correlation search and measure sizes",
                                "sofi.sofi_img_jitter",
                                "40,40,65,65");
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "xcorr");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);

    /* --comb_meth */
    p = cpl_parameter_new_value("sofi.sofi_img_jitter.comb_meth", 
            CPL_TYPE_STRING, "union / inter / first", "sofi.sofi_img_jitter",
            "union");
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "comb_meth");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
 
    /* --rej */
    p = cpl_parameter_new_value("sofi.sofi_img_jitter.rej",
                                CPL_TYPE_STRING,
                                "Low and high number of rejected values",
                                "sofi.sofi_img_jitter",
                                "2,2");
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "rej");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);

    /* --row_med */
    p = cpl_parameter_new_value("sofi.sofi_img_jitter.row_med", CPL_TYPE_BOOL,
            "flag to subtract the median of each row", "sofi.sofi_img_jitter",
            TRUE);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "row_med");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);

    /* Return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int sofi_img_jitter_exec(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    return sofi_img_jitter(recipe->parameters, recipe->frames);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int sofi_img_jitter_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    cpl_parameterlist_delete(recipe->parameters);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the command line options and execute the data reduction
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int sofi_img_jitter(
        cpl_parameterlist   *   parlist, 
        cpl_frameset        *   framelist)
{
    const char      *   sval;
    cpl_parameter   *   par;
    const char      *   flat;
    const char      *   dark;
    cpl_frameset    *   objframes;
    cpl_frameset    *   skyframes;
    cpl_image       **  combined;
    cpl_table       *   objs_stats;
    cpl_vector      *   sky_bg;

    /* Initialise */
    par = NULL;
    sofi_img_jitter_config.pixscale = -1.0;
    sofi_img_jitter_config.dit = -1.0;
    sofi_img_jitter_config.iq = -1.0;
    sofi_img_jitter_config.nbobjs = -1;
    sofi_img_jitter_config.fwhm_pix = -1.0;
    sofi_img_jitter_config.fwhm_arcsec = -1.0;
    sofi_img_jitter_config.fwhm_mode = -1.0;
    sofi_img_jitter_config.offsets = NULL;
    sofi_img_jitter_config.objects = NULL;
    sofi_img_jitter_config.nb_obj_frames = 0;
    sofi_img_jitter_config.nb_rej_frames = 0;
    sofi_img_jitter_config.nb_sky_frames = 0;

    /* Retrieve input parameters */
    /* Offsets */
    par = cpl_parameterlist_find(parlist, "sofi.sofi_img_jitter.offsets");
    sofi_img_jitter_config.offsets = cpl_parameter_get_string(par);
    /* Objects */
    par = cpl_parameterlist_find(parlist, "sofi.sofi_img_jitter.objects");
    sofi_img_jitter_config.objects = cpl_parameter_get_string(par);
    /* Cross-talk */
    par = cpl_parameterlist_find(parlist, "sofi.sofi_img_jitter.crosstalk");
    sofi_img_jitter_config.crosstalk = cpl_parameter_get_bool(par);
    /* Rejection parameters for sky filtering */
    par = cpl_parameterlist_find(parlist, "sofi.sofi_img_jitter.sky_par");
    sval = cpl_parameter_get_string(par);
    if (sscanf(sval, "%d,%d,%d,%d",
                    &sofi_img_jitter_config.sky_minnb,
                    &sofi_img_jitter_config.sky_halfw,
                    &sofi_img_jitter_config.sky_rejmin,
                    &sofi_img_jitter_config.sky_rejmax)!=4) {
        return CPL_ERROR_UNSPECIFIED;
    }
    /* Cross correlation windows parameters */
    par = cpl_parameterlist_find(parlist, "sofi.sofi_img_jitter.xcorr");
    sval = cpl_parameter_get_string(par);
    if (sscanf(sval, "%d,%d,%d,%d",
                    &sofi_img_jitter_config.sx,
                    &sofi_img_jitter_config.sy,
                    &sofi_img_jitter_config.mx,
                    &sofi_img_jitter_config.my)!=4) {
        return CPL_ERROR_UNSPECIFIED;
    }
    /* --comb_meth */
    par = cpl_parameterlist_find(parlist, "sofi.sofi_img_jitter.comb_meth");
    sval = cpl_parameter_get_string(par);
    if (!strcmp(sval, "union")) 
        sofi_img_jitter_config.comb_meth = CPL_GEOM_UNION;
    else if (!strcmp(sval, "inter"))
        sofi_img_jitter_config.comb_meth = CPL_GEOM_INTERSECT;
    else if (!strcmp(sval, "first"))
        sofi_img_jitter_config.comb_meth = CPL_GEOM_FIRST;
    else {
        cpl_msg_error(cpl_func, "Invalid combine method specified");
        return CPL_ERROR_UNSPECIFIED;
    }
    
     /* Number of rejected values in stacking */
    par = cpl_parameterlist_find(parlist, "sofi.sofi_img_jitter.rej");
    sval = cpl_parameter_get_string(par);
    if (sscanf(sval, "%d,%d",
                    &sofi_img_jitter_config.rej_low,
                    &sofi_img_jitter_config.rej_high)!=2) {
        return CPL_ERROR_UNSPECIFIED;
    }
    /* Row median */
    par = cpl_parameterlist_find(parlist, "sofi.sofi_img_jitter.row_med");
    sofi_img_jitter_config.row_med = cpl_parameter_get_bool(par);

    /* Identify the RAW and CALIB frames in the input frameset */
    if (sofi_dfs_set_groups(framelist)) {
        cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames");
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Retrieve calibration data */
    flat   = sofi_extract_filename(framelist, SOFI_CALIB_FLAT);
    dark   = sofi_extract_filename(framelist, SOFI_CALIB_DARK);

    /* Retrieve raw frames */
    if ((objframes = sofi_extract_frameset(framelist,
                    SOFI_IMG_JITTER_OBJ_RAW)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot find objs frames in the input list");
        return CPL_ERROR_UNSPECIFIED;
    }
    skyframes = sofi_extract_frameset(framelist, SOFI_IMG_JITTER_SKY_RAW);

    /* Apply the reduction */
    cpl_msg_info(cpl_func, "Apply the data recombination");
    cpl_msg_indent_more();
    if ((combined = sofi_img_jitter_reduce(objframes, skyframes, flat,
                    dark, NULL, &sky_bg)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot recombine the data");
        cpl_frameset_delete(objframes);
        if (skyframes) cpl_frameset_delete(skyframes);
        cpl_msg_indent_less();
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_frameset_delete(objframes);
    if (skyframes) cpl_frameset_delete(skyframes);
    cpl_msg_indent_less();
   
    /* Compute QC parameters from the combined image */
    cpl_msg_info(cpl_func, "Compute QC parameters from the combined image");
    cpl_msg_indent_more();
    if ((objs_stats = sofi_img_jitter_qc(combined[0])) == NULL) {
        cpl_msg_warning(cpl_func, "Cannot compute all parameters");
    }
    cpl_msg_indent_less();
    
    /* Save the products */
    cpl_msg_info(cpl_func, "Save the products");
    cpl_msg_indent_more();
    if (sofi_img_jitter_save(combined[0], objs_stats, sky_bg, parlist,
                framelist) == -1){
        cpl_msg_error(cpl_func, "Cannot save the products");
        cpl_image_delete(combined[0]);
        cpl_image_delete(combined[1]);
        cpl_free(combined);
        if (objs_stats) cpl_table_delete(objs_stats);
        if (sky_bg) cpl_vector_delete(sky_bg);
        cpl_msg_indent_less();
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_msg_indent_less();
    
    /* Return */
    cpl_image_delete(combined[0]);
    cpl_image_delete(combined[1]);
    cpl_free(combined);
    if (objs_stats) cpl_table_delete(objs_stats);
        if (sky_bg) cpl_vector_delete(sky_bg);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    obj         the objects frames
  @param    sky         the sky frames or NULL
  @param    flat        the flat field or NULL
  @param    dark        the dark or NULL
  @param    bpm         the bad pixels map or NULL
  @param    skybg       the computed sky background values
  @return   the combined image and the contribution map  or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_image ** sofi_img_jitter_reduce(
        cpl_frameset    *   obj,
        cpl_frameset    *   sky,
        const char      *   flat,
        const char      *   dark,
        const char      *   bpm,
        cpl_vector      **  skybg)
{
    cpl_imagelist   **  in;
    cpl_image       **  combined;

    /* Initialise */
    *skybg = NULL;
    
    /* Load the input data */
    cpl_msg_info(cpl_func, "Load the input data");
    cpl_msg_indent_more();
    if ((in = sofi_img_jitter_load(obj, sky)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot load input data");
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_msg_indent_less();

    /* Apply the cross-talk correction */
    if (sofi_img_jitter_config.crosstalk) {
        cpl_msg_info(cpl_func, "Apply the cross-talk correction");
        cpl_msg_indent_more();
        if (sofi_correct_crosstalk_list(in[0]) == -1) {
            cpl_msg_error(cpl_func, "Cannot correct for Cross-talk");
            cpl_imagelist_delete(in[0]);
            if (in[1]) cpl_imagelist_delete(in[1]);
            cpl_free(in);
            cpl_msg_indent_less();
            return NULL;
        }
        if (in[1]) {
            if (sofi_correct_crosstalk_list(in[1]) == -1) {
                cpl_msg_error(cpl_func, "Cannot correct for Cross-talk");
                cpl_imagelist_delete(in[0]);
                if (in[1]) cpl_imagelist_delete(in[1]);
                cpl_free(in);
                cpl_msg_indent_less();
                return NULL;
            }
        }
        cpl_msg_indent_less();
    }
    
    /* Apply the calibrations */
    if (flat || dark || bpm) {
        cpl_msg_info(cpl_func, "Apply the calibrations");
        cpl_msg_indent_more();
        if (irplib_flat_dark_bpm_calib(in[0], flat, dark, bpm) == -1) {
            /* Calibrate the objects */
            cpl_msg_error(cpl_func, "Cannot calibrate the objects");
            cpl_imagelist_delete(in[0]);
            if (in[1]) cpl_imagelist_delete(in[1]);
            cpl_free(in);
            cpl_msg_indent_less();
            return NULL;
        }
        if (in[1]) {
            if (irplib_flat_dark_bpm_calib(in[1], flat, dark, bpm) == -1) {
                /* Calibrate the sky */
                cpl_msg_error(cpl_func, "Cannot calibrate the sky");
                cpl_imagelist_delete(in[0]);
                if (in[1]) cpl_imagelist_delete(in[1]);
                cpl_free(in);
                cpl_msg_indent_less();
                return NULL;
            }
        }
        cpl_msg_indent_less();
    }

    /* Apply the sky correction */
    cpl_msg_info(cpl_func, "Sky estimation and correction");
    cpl_msg_indent_more();
    if ((*skybg = sofi_img_jitter_sky(&(in[0]), in[1])) == NULL) {
        cpl_msg_error(cpl_func, "Cannot estimate the sky");
        cpl_imagelist_delete(in[0]);
        if (in[1]) cpl_imagelist_delete(in[1]);
        cpl_free(in);
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_msg_indent_less();
    sofi_img_jitter_config.nb_obj_frames = cpl_imagelist_get_size(in[0]);
    if (in[1] != NULL) 
        sofi_img_jitter_config.nb_sky_frames = cpl_imagelist_get_size(in[1]);
    if (in[1]) cpl_imagelist_delete(in[1]);
    in[1] = NULL;
    
    /* Apply the shift and add */
    cpl_msg_info(cpl_func, "Shift and add");
    cpl_msg_indent_more();
    combined = sofi_img_jitter_saa(in[0], obj);
    if (combined == NULL) {
        cpl_msg_error(cpl_func, "Cannot apply the shift and add");
        cpl_imagelist_delete(in[0]);
        cpl_free(in);
        if (*skybg != NULL) cpl_vector_delete(*skybg);
        *skybg = NULL;
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_imagelist_delete(in[0]);
    cpl_free(in);
    cpl_msg_indent_less();
     
    /* Post processing on the combined image */
    if (sofi_img_jitter_config.row_med) {
        cpl_msg_info(cpl_func, "Subtract the median from each row");
        sofi_img_jitter_sub_row_median(combined[0]);
    }
    
    return combined;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Load the input data (sky and object) 
  @param    obj     the objects frames
  @param    sky     the sky frames or NULL
  @return   A newly allocated pointer to 2 newly allocated images sets, or NULL
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist ** sofi_img_jitter_load(
        cpl_frameset    *   obj,
        cpl_frameset    *   sky)
{
    cpl_imagelist       **  isets;
    cpl_frame           *   frame;
    cpl_propertylist    *   plist;
    
    /* Test entries */
    if (obj == NULL) return NULL;

    /* Check if ok */
    if (cpl_error_get_code()) return NULL;
   
    /* Get the frame type pixscale and DIT value */
    frame = cpl_frameset_get_position(obj, 0);
    plist=cpl_propertylist_load(cpl_frame_get_filename(frame), 0);
    sofi_img_jitter_config.pixscale = sofi_pfits_get_pixscale(plist);
    sofi_img_jitter_config.dit = sofi_pfits_get_dit(plist);
    cpl_propertylist_delete(plist);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Missing keyword in FITS header");
        return NULL;
    }

    /* Allocate the image sets */
    isets = cpl_malloc(2 * sizeof(cpl_imagelist *));
    isets[0] = cpl_imagelist_load_frameset(obj, CPL_TYPE_FLOAT, 1, 0);
    if (sky != NULL) 
        isets[1] = cpl_imagelist_load_frameset(sky, CPL_TYPE_FLOAT, 1, 0);
    else isets[1] = NULL;
    
    /* Check if at least the objects are there */
    if (isets[0] == NULL) {
        cpl_msg_error(cpl_func, "The objects frames could not be loaded");
        if (isets[1] != NULL) cpl_imagelist_delete(isets[1]);
        cpl_free(isets);
        return NULL;
    }
    /* Return */
    return isets;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Estimate the sky signal and correct the images.
  @param    objs    the images with objects
  @param    skys    the images with the sky or NULL
  @return   the background values in a vector or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * sofi_img_jitter_sky(
        cpl_imagelist   **  objs,
        cpl_imagelist   *   skys)
{
    cpl_image       *   sky;
    cpl_vector      *   bg;
    int                 sky_method;
    int                 nframes, nskys;
    double              median;
    cpl_image       *   cur_ima;
    int                 i;

    /* Initialise */
    nframes = cpl_imagelist_get_size(*objs);
    bg = NULL;

    /* Decide the sky method */
    if (sofi_img_jitter_config.sky_minnb > nframes) sky_method = 1;
    else sky_method = 2;

    /* Compute the sky frame */
    if (skys != NULL) {
        cpl_msg_info(cpl_func, "Median of sky images");
        /* Get the median of the sky images */
        nskys = cpl_imagelist_get_size(skys);
        bg = cpl_vector_new(nskys);
        for (i=0 ; i<nskys ; i++) {
            median = cpl_image_get_median(cpl_imagelist_get(skys, i));
            cpl_vector_set(bg, i, median);
        }
        /* Use sky images */
        if ((sky = cpl_imagelist_collapse_median_create(skys)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot compute the median of sky images");
            cpl_vector_delete(bg);
            return NULL;
        }
        /* Correct the objects images  */
        if (cpl_imagelist_subtract_image(*objs, sky) != CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "Cannot correct the obj images from the sky");
            cpl_image_delete(sky);
            cpl_vector_delete(bg);
            return NULL;
        }
        cpl_image_delete(sky);
        /* Normalise the object planes */
        for (i=0 ; i<nframes ; i++) {
            cur_ima = cpl_imagelist_get(*objs, i);
            median = cpl_image_get_median(cur_ima);
            cpl_image_subtract_scalar(cur_ima, median);
        }
    } else if (sky_method == 1) {
        cpl_msg_info(cpl_func, "Median of object images");
         /* Use objs images */
        if ((sky = cpl_imagelist_collapse_median_create(*objs)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot compute the median of obj images");
            return NULL;
        }
        /* Correct the objects images  */
        if (cpl_imagelist_subtract_image(*objs, sky) != CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "Cannot correct the obj images from the sky");
            cpl_image_delete(sky);
            return NULL;
        }
        /* Create a 1 value vector */
        bg = cpl_vector_new(1);
        cpl_vector_set(bg, 0, cpl_image_get_median(sky));
        cpl_image_delete(sky);
        /* Normalise the object planes */
        for (i=0 ; i<nframes ; i++) {
            cur_ima = cpl_imagelist_get(*objs, i);
            median = cpl_image_get_median(cur_ima);
            cpl_image_subtract_scalar(cur_ima, median);
        }
    } else if (sky_method == 2) {
        cpl_msg_info(cpl_func, "Running filter on object images");
        /* Use objects images */
        if ((bg = sofi_img_jitter_sky_running(objs)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot apply the running filter for the sky");
            return NULL;
        }
    }
    /* Free and return */
    return bg;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Apply the running filter to subtract the sky
  @param    in          the images to correct
  @return   the vector containing the background values or NULL in error case

  This function performs a 3d rejection filter on the input cube.
  Each time line is extracted on +/-halfw. On this line of sight, all
  pixels are first normalized by subtracting the median value of the
  plane they belong to. The central, highest and lowest values are then
  removed and the rest is averaged to yield a value which is subtracted
  from the initial input pixel.

  The background array is returned.
  The array will contain the average value which has been subtracted
  to pixels in each plane. It is usually a good indicator of the
  average subtracted background value if you use this filter to subtract
  an infrared sky background.
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * sofi_img_jitter_sky_running(cpl_imagelist ** in) 
{
    int                 rejmin, rejmax, halfw;
    cpl_imagelist   *   filtres;
    int                 ni, nx, ny, pos;
    cpl_vector      *   medians;
    cpl_image       *   cur_ima;
    float           *   pcur_ima;
    cpl_image       *   tmp_ima;
    float           *   ptmp_ima;
    cpl_vector      *   localwin;
    int                 fr_p, to_p, n_curp;
    cpl_vector      *   bg;
    double              bg_val, out;
    float               one_med;
    int                 i, j, k;

    /* Test entries */
    if (in==NULL || *in == NULL) return NULL;

    /* Initialise */
    rejmin = sofi_img_jitter_config.sky_rejmin;
    rejmax = sofi_img_jitter_config.sky_rejmax;
    halfw  = sofi_img_jitter_config.sky_halfw;
    ni = cpl_imagelist_get_size(*in);
    cur_ima = cpl_imagelist_get(*in, 0);
    nx = cpl_image_get_size_x(cur_ima);
    ny = cpl_image_get_size_y(cur_ima);
    
    /* Tests on validity of rejection parameters */
    if (((rejmin+rejmax)>=halfw) || (halfw<1) || (rejmin<0) || (rejmax<0)) {
        cpl_msg_error(cpl_func, "cannot run filter with rej parms %d (%d-%d)",
                halfw, rejmin, rejmax);
        return NULL;
    }   
    /* Pre-compute median value in each plane */
    medians = cpl_vector_new(ni);
    for (i=0 ; i<ni ; i++) {
        cur_ima = cpl_imagelist_get(*in, i);
        cpl_vector_set(medians, i, cpl_image_get_median(cur_ima));
    }
    /* Allocate output cube */
    filtres = cpl_imagelist_new();

    /* Allocate output bg */
    bg = cpl_vector_new(ni);
    
    /* Main loop over input planes */
    for (k=0 ; k<ni ; k++) {
        /* Allocate output plane */
        tmp_ima = cpl_image_new(nx, ny, CPL_TYPE_FLOAT);
        ptmp_ima = cpl_image_get_data_float(tmp_ima);
        /* Compute border indices */
        fr_p = k - halfw;
        to_p = k + halfw;
        if (fr_p<0) fr_p=0;
        if (to_p>(ni-1)) to_p=ni-1;

        /* Number of valid planes to consider after edge effects */
        n_curp = to_p - fr_p;

        /* Allocate local window */
        localwin = cpl_vector_new(n_curp);
        
        bg_val = 0.0;
        /* Loop over all pixels */
        for (pos=0 ; pos<nx*ny ; pos++) {
            /* Fill up local window */
            j=0;
            for (i=fr_p ; i<=to_p ; i++) {
                if (i!=k) {
                    cur_ima = cpl_imagelist_get(*in, i);
                    pcur_ima = cpl_image_get_data_float(cur_ima);
                    cpl_vector_set(localwin, j,
                            (double)pcur_ima[pos]-cpl_vector_get(medians, i));
                    j++;
                }
            }
            /* Sort window */
            cpl_vector_sort(localwin, 1);
            /* Reject min and max, accumulate other pixels */
            out = 0.0;
            for (i=rejmin ; i<(n_curp-rejmax) ; i++) {
                out += cpl_vector_get(localwin, i);
            }
            /* Take the mean */
            out /= (double)(n_curp - rejmin - rejmax);
            /* Assign value */
            cur_ima = cpl_imagelist_get(*in, k);
            pcur_ima = cpl_image_get_data_float(cur_ima);
            ptmp_ima[pos] = pcur_ima[pos] - 
                (float)(out + cpl_vector_get(medians, k));

            bg_val += (out+cpl_vector_get(medians, k));
        }
        cpl_vector_delete(localwin);
        cpl_vector_set(bg, k, bg_val/(nx*ny));
        cpl_imagelist_set(filtres, tmp_ima, k);
    }
    cpl_imagelist_delete(*in);
    cpl_vector_delete(medians);

    /* Subtract median from each frame */
    for (i=0 ; i<ni ; i++) {
        cur_ima = cpl_imagelist_get(filtres, i);
        one_med = cpl_image_get_median(cur_ima);
        cpl_image_subtract_scalar(cur_ima, one_med);
    }
    *in = filtres;
    return bg;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Apply the shift and add in jitter mode
  @param    in          the images to stack
  @param    objframes   the list of objects frames
  @return   the combined image and the contribution map or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_image ** sofi_img_jitter_saa(
        cpl_imagelist   *   in,
        cpl_frameset    *   objframes)
{
    cpl_frame           *   frame;
    cpl_propertylist    *   plist;
    cpl_bivector        *   offsets_est;
    double              *   offsets_est_x;
    double              *   offsets_est_y;
    cpl_bivector        *   objs;
    double              *   objs_x;
    double              *   objs_y;
    cpl_apertures       *   aperts;
    cpl_image           **  combined;
    cpl_vector          *   thresh_vect;
    cpl_image           *   diff;
    int                     nfiles;
    int                     i;

    /* Get the number of images */
    nfiles = cpl_imagelist_get_size(in);
    if (cpl_frameset_get_size(objframes) != nfiles) {
        cpl_msg_error(cpl_func, "Invalid input objects sizes");
        return NULL;
    }
    
    /* Get the offsets estimation of each input file pair */
    cpl_msg_info(cpl_func, "Get the offsets estimation");
    offsets_est = NULL;
    if (sofi_img_jitter_config.offsets &&
            sofi_img_jitter_config.offsets[0] != (char)0) {
        /* A file has been provided on the command line */
        offsets_est = cpl_bivector_read(sofi_img_jitter_config.offsets);
        if ((offsets_est==NULL)||(cpl_bivector_get_size(offsets_est)!=nfiles)) {
            cpl_msg_error(cpl_func, "Cannot get offsets from %s", 
                    sofi_img_jitter_config.offsets);
            return NULL;
        }
    } else {
        /* Get the offsets from the header */
        offsets_est = cpl_bivector_new(nfiles);
        offsets_est_x = cpl_bivector_get_x_data(offsets_est);
        offsets_est_y = cpl_bivector_get_y_data(offsets_est);
        for (i=0 ; i<nfiles ; i++) {
            if (cpl_error_get_code()) {
                cpl_bivector_delete(offsets_est);
                cpl_ensure(0, cpl_error_get_code(), NULL);
            }
            /* X and Y offsets */
            frame = cpl_frameset_get_position(objframes, i);
            plist=cpl_propertylist_load(cpl_frame_get_filename(frame),0);
            offsets_est_x[i] = -1.0 * sofi_pfits_get_cumoffsetx(plist);
            offsets_est_y[i] = -1.0 * sofi_pfits_get_cumoffsety(plist);
            cpl_propertylist_delete(plist);
            if (cpl_error_get_code()) {
                cpl_msg_warning(cpl_func, "Cannot get offsets from header");
                cpl_bivector_delete(offsets_est);
                cpl_ensure(0, cpl_error_get_code(), NULL);
            }
        }
        /* Subtract the first offset to all offsets */
        for (i=1 ; i<nfiles ; i++) {
            offsets_est_x[i] -= offsets_est_x[0];
            offsets_est_y[i] -= offsets_est_y[0];
        }
        offsets_est_x[0] = offsets_est_y[0] = 0.00;
    }

    /* Read the provided objects file if provided */
    objs = NULL;
    if (sofi_img_jitter_config.objects &&
            sofi_img_jitter_config.objects[0] != (char)0) {
        cpl_msg_info(cpl_func, "Get the user provided correlation objects");
        /* A file has been provided on the command line */
        objs = cpl_bivector_read(sofi_img_jitter_config.objects);
        if (objs==NULL) {
            cpl_msg_error(cpl_func, "Cannot get objects from %s",
                    sofi_img_jitter_config.objects);
            if (offsets_est) cpl_bivector_delete(offsets_est);
            return NULL;
        }
    }

    /* Get a correlation point from the difference of the first images */
    if (objs == NULL) {
        cpl_msg_info(cpl_func, "Get a cross-correlation point");
        thresh_vect = cpl_vector_new(4);
        cpl_vector_set(thresh_vect, 0, 5.0);
        cpl_vector_set(thresh_vect, 1, 2.0);
        cpl_vector_set(thresh_vect, 2, 1.0);
        cpl_vector_set(thresh_vect, 3, 0.5);
        diff = cpl_image_subtract_create(cpl_imagelist_get(in, 0),
                cpl_imagelist_get(in, 1));
        if ((aperts = cpl_apertures_extract_window(diff, thresh_vect, 
                        200, 200, 800, 800, NULL)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot find any cross-correlation point");
            if (offsets_est) cpl_bivector_delete(offsets_est);
            cpl_vector_delete(thresh_vect);
            cpl_image_delete(diff);
            return NULL;
        }
        cpl_image_delete(diff);
        cpl_vector_delete(thresh_vect);
        cpl_apertures_sort_by_npix(aperts);
        objs = cpl_bivector_new(1);
        objs_x = cpl_bivector_get_x_data(objs);
        objs_y = cpl_bivector_get_y_data(objs);
        objs_x[0] = cpl_apertures_get_pos_x(aperts, 1);
        objs_y[0] = cpl_apertures_get_pos_y(aperts, 1);
        cpl_apertures_delete(aperts);
        if (objs == NULL) {
            cpl_msg_error(cpl_func, "Cannot find any cross-correlation point");
            if (offsets_est) cpl_bivector_delete(offsets_est);
            return NULL;
        }
        cpl_msg_info(cpl_func, "Correlation point: %g %g\n", objs_x[0], objs_y[0]);
    }
    
    /* Recombine the images */
    cpl_msg_info(cpl_func, "Recombine the images set");
    cpl_msg_indent_more();
    if ((combined = cpl_geom_img_offset_combine(in, offsets_est, 1, objs,
                    NULL, NULL,
                    sofi_img_jitter_config.sx, 
                    sofi_img_jitter_config.sy,
                    sofi_img_jitter_config.mx, 
                    sofi_img_jitter_config.my,
                    sofi_img_jitter_config.rej_low,
                    sofi_img_jitter_config.rej_high,
                    sofi_img_jitter_config.comb_meth)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot recombine the images");
        if (offsets_est) cpl_bivector_delete(offsets_est);
        cpl_bivector_delete(objs);
        cpl_msg_indent_less();
        return NULL;
    }
    /* Update QC params */
    i = (int)(cpl_image_get_max(combined[1]));
    sofi_img_jitter_config.nb_rej_frames = 
        sofi_img_jitter_config.nb_obj_frames - i;
    sofi_img_jitter_config.nb_obj_frames = i;
    cpl_msg_indent_less();

    /* Free and return */
    if (offsets_est) cpl_bivector_delete(offsets_est);
    cpl_bivector_delete(objs);
    return combined;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Subtract from each row the median of the row
  @param    combined    the combined image 
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int sofi_img_jitter_sub_row_median(cpl_image * in)
{
    float           *   pin;
    int                 nx, ny;
    double              median;
    int                 i, j;

    /* Test entries */
    if (in == NULL) return CPL_ERROR_UNSPECIFIED;

    /* Initialise */
    nx = cpl_image_get_size_x(in);
    ny = cpl_image_get_size_y(in);
    pin = cpl_image_get_data_float(in);
 
    for (i=0 ; i<ny ; i++) {
        median = cpl_image_get_median_window(in, 1, i+1, nx, i+1);
        for (j=0 ; j<nx ; j++) {
            pin[j+i*nx] -= (float)median;
        }
    }
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute some QC parameters from the combined image
  @param    combined    the combined image produced
  @return   a newly allocated table with the stars stats or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_table * sofi_img_jitter_qc(cpl_image * combined)
{
    cpl_apertures   *   aperts;
    cpl_bivector    *   fwhms;
    double          *   fwhms_x;
    double          *   fwhms_y;
    cpl_vector      *   fwhms_good;
    double          *   fwhms_good_data;
    int                 nb_val, nb_good;
    double              f_min, f_max, fr, fx, fy;
    cpl_vector      *   thresh_vec;
    cpl_table       *   out_tab;
    int                 nb_objs;
    int                 i, j;
    
    /* Initialise */
    double              seeing_min_arcsec = 0.1;
    double              seeing_max_arcsec = 5.0;
    double              seeing_fwhm_var   = 0.2;

    /* Check entries */
    if (combined == NULL) return NULL;
    
    /* Create the vector for the detection thresholds */
    thresh_vec = cpl_vector_new(4);
    cpl_vector_set(thresh_vec, 0, 5.0);
    cpl_vector_set(thresh_vec, 1, 2.0);
    cpl_vector_set(thresh_vec, 2, 1.0);
    cpl_vector_set(thresh_vec, 3, 0.5);

    /* Detect apertures */
    if ((aperts = cpl_apertures_extract(combined, thresh_vec, NULL)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot detect any aperture");
        cpl_vector_delete(thresh_vec);
        return NULL;
    }
    cpl_vector_delete(thresh_vec);

    /* Number of detected objects */
    nb_objs = cpl_apertures_get_size(aperts);
    sofi_img_jitter_config.nbobjs = nb_objs;

    /* Compute the FHWM of the detected apertures */
    CPL_DIAG_PRAGMA_PUSH_IGN(-Wdeprecated-declarations);
    fwhms = cpl_apertures_get_fwhm(combined, aperts);
    CPL_DIAG_PRAGMA_POP;

    if (fwhms == NULL) {
        cpl_msg_error(cpl_func, "Cannot compute the FWHMs");
        cpl_apertures_delete(aperts);
        return NULL;
    }

    /* Access the data */
    nb_val = cpl_bivector_get_size(fwhms);
    fwhms_x = cpl_bivector_get_x_data(fwhms);
    fwhms_y = cpl_bivector_get_y_data(fwhms);

    /* Store everything in the output table */
    out_tab = cpl_table_new(nb_objs);
    cpl_table_new_column(out_tab, "POS_X", CPL_TYPE_DOUBLE);
    cpl_table_new_column(out_tab, "POS_Y", CPL_TYPE_DOUBLE);
    cpl_table_new_column(out_tab, "FWHM_X", CPL_TYPE_DOUBLE);
    cpl_table_new_column(out_tab, "FWHM_Y", CPL_TYPE_DOUBLE);
    cpl_table_new_column(out_tab, "FLUX", CPL_TYPE_DOUBLE);
    for (i=0 ; i<nb_objs ; i++) {
        cpl_table_set_double(out_tab, "POS_X", i, 
                cpl_apertures_get_centroid_x(aperts, i+1));
        cpl_table_set_double(out_tab, "POS_Y", i, 
                cpl_apertures_get_centroid_y(aperts, i+1));
        cpl_table_set_double(out_tab, "FWHM_X", i, fwhms_x[i]);
        cpl_table_set_double(out_tab, "FWHM_Y", i, fwhms_y[i]);
        cpl_table_set_double(out_tab, "FLUX", i, 
                cpl_apertures_get_flux(aperts, i+1));
    }
    cpl_apertures_delete(aperts);

    /* Get the number of good values */
    nb_good = 0;
    for (i=0 ; i<nb_val ; i++) {
        if ((fwhms_x[i] > 0.0) && (fwhms_y[i] > 0.0)) nb_good++;
    }
    if (nb_good == 0) {
        cpl_table_delete(out_tab);
        cpl_bivector_delete(fwhms);
        (void)cpl_error_set(cpl_func, CPL_ERROR_DATA_NOT_FOUND);
        return NULL;
    }
    
    /* Get the good values */
    fwhms_good = cpl_vector_new(nb_good);
    fwhms_good_data = cpl_vector_get_data(fwhms_good);
    j=0;
    for (i=0 ; i<nb_val ; i++) {
        if ((fwhms_x[i] > 0.0) && (fwhms_y[i] > 0.0)) {
            fwhms_good_data[j] = (fwhms_x[i]+fwhms_y[i])/2.0;
            j++;
        }
    }
   
    /* Compute the fwhm */
    if (nb_good < 3) {
        /* Too few values to compute the median */
        sofi_img_jitter_config.fwhm_pix = fwhms_good_data[0];
    } else {
        /* Compute the median */
        sofi_img_jitter_config.fwhm_pix = cpl_vector_get_median_const(fwhms_good);
    }
    sofi_img_jitter_config.fwhm_arcsec = 
        sofi_img_jitter_config.fwhm_pix * sofi_img_jitter_config.pixscale;

    /* Compute the mode of the FWHMs */
    if (nb_good > 5) {
        sofi_img_jitter_config.fwhm_mode = 
            sofi_img_jitter_get_mode(fwhms_good);
    }
    cpl_vector_delete(fwhms_good);
    
    /* IQ is the median of the (fwhm_x+fwhm_y/2) of the good stars */
    /* Compute f_min and f_max */
    f_min = seeing_min_arcsec / sofi_img_jitter_config.pixscale;
    f_max = seeing_max_arcsec / sofi_img_jitter_config.pixscale;

    /* Get the number of good values */
    nb_good = 0;
    for (i=0 ; i<nb_val ; i++) {
        fx = fwhms_x[i];
        fy = fwhms_y[i];
        fr = 2.0 * fabs(fx-fy) / (fx+fy);
        if ((fx > f_min) && (fx < f_max) && (fy > f_min) && (fy < f_max) &&
            (fr < seeing_fwhm_var)) nb_good++;
    }
    if (nb_good == 0) {
        cpl_table_delete(out_tab);
        cpl_bivector_delete(fwhms);
        (void)cpl_error_set(cpl_func, CPL_ERROR_DATA_NOT_FOUND);
        return NULL;
    }

    /* Get the good values */
    fwhms_good = cpl_vector_new(nb_good);
    fwhms_good_data = cpl_vector_get_data(fwhms_good);
    j=0;
    for (i=0 ; i<nb_val ; i++) {
        fx = fwhms_x[i];
        fy = fwhms_y[i];
        fr = 2.0 * fabs(fx-fy) / (fx+fy);
        if ((fx > f_min) && (fx < f_max) && (fy > f_min) && (fy < f_max) &&
            (fr < seeing_fwhm_var)) {
            fwhms_good_data[j] = (fx + fy)/2.0;
            j++;
        }
    }
    cpl_bivector_delete(fwhms);
    
    /* Compute the fwhm */
    if (nb_good < 3) {
        /* Too few values to compute the median */
        sofi_img_jitter_config.iq = fwhms_good_data[0];
    } else {
        /* Compute the median */
        sofi_img_jitter_config.iq = cpl_vector_get_median_const(fwhms_good);
    }
    cpl_vector_delete(fwhms_good);
    sofi_img_jitter_config.iq *= sofi_img_jitter_config.pixscale;

    return out_tab;
}
        
/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the histogram mode
  @param    vec         the vector 
  @return   the mode or -1 in error case
 */
/*----------------------------------------------------------------------------*/
static double sofi_img_jitter_get_mode(cpl_vector * vec)
{
    int                 nb;
    int                 nbins;
    double              min, max;
    double              bin_size;
    cpl_bivector    *   hist;
    cpl_vector      *   hist_x;
    cpl_vector      *   hist_y;
    double              cur_val;
    int                 cur_bin;
    double              max_val;
    int                 max_bin;
    double              mode;
    int                 i;

    /* Test entries  */
    if (vec == NULL) return -1.0;
    
    /* Initialise */
    nb = cpl_vector_get_size(vec);

    /* Create the histogram */
    nbins = 10;
    min = cpl_vector_get_min(vec);
    max = cpl_vector_get_max(vec);
    bin_size = (max-min)/nbins;
    hist = cpl_bivector_new(nbins);
    hist_x = cpl_bivector_get_x(hist);
    hist_y = cpl_bivector_get_y(hist);
    cpl_vector_fill(hist_x, 0.0);
    cpl_vector_fill(hist_y, 0.0);
    for (i=0 ; i<nbins ; i++) {
        cpl_vector_set(hist_x, i, min + i * bin_size);
    }
    for (i=0 ; i<nb ; i++) {
        cur_val = cpl_vector_get(vec, i);
        cur_bin = (int)((cur_val - min) / bin_size);
        if (cur_bin >= nbins) cur_bin -= 1.0;
        cur_val = cpl_vector_get(hist_y, cur_bin);
        cur_val += 1.0;
        cpl_vector_set(hist_y, cur_bin, cur_val);
    }
    
    /* Get the mode of the histogram */
    max_val = cpl_vector_get(hist_y, 0);
    max_bin = 0;
    for (i=0 ; i<nbins ; i++) {
        cur_val = cpl_vector_get(hist_y, i);
        if (cur_val > max_val) {
            max_val = cur_val;
            max_bin = i;
        }
    }
    mode = cpl_vector_get(hist_x, max_bin);
    cpl_bivector_delete(hist);
    return mode;
}
        
/*----------------------------------------------------------------------------*/
/**
  @brief    Save the jitter recipe products on disk
  @param    combined    the combined image produced
  @param    objs_stats  the table with the detected objects statistics or NULL
  @param    sky_bg      the vector with the sky background values or NULL
  @param    parlist     the input list of parameters
  @param    set         the input frame set
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int sofi_img_jitter_save(
        cpl_image           *   combined,
        cpl_table           *   objs_stats,
        cpl_vector          *   sky_bg,
        cpl_parameterlist   *   parlist,
        cpl_frameset        *   set)
{
    const cpl_frame     *   ref_frame;
    cpl_propertylist    *   plist;
    cpl_propertylist    *   paflist;
    cpl_propertylist    *   qclist;
    double                  pscale, dit, bg_mean, bg_stdev, bg_instmag;
    cpl_table           *   sky_bg_tab;
    int                     i;
 
    /* Get the QC params in qclist */
    qclist = cpl_propertylist_new();

    if (sky_bg != NULL) {
        pscale = sofi_img_jitter_config.pixscale;
        dit = sofi_img_jitter_config.dit;
        bg_mean = cpl_vector_get_mean(sky_bg);
        if (cpl_vector_get_size(sky_bg) < 2) bg_stdev = 0;
        else bg_stdev = cpl_vector_get_stdev(sky_bg);
        bg_instmag = -2.5 * log(bg_mean/(pscale*pscale*dit));
        cpl_propertylist_append_double(qclist, "ESO QC BACKGD MEAN", bg_mean);
        cpl_propertylist_append_double(qclist, "ESO QC BACKGD STDEV", bg_stdev);
        cpl_propertylist_append_double(qclist, "ESO QC BACKGD INSTMAG", 
                bg_instmag);
    }
    cpl_propertylist_append_int(qclist, "ESO QC NBOBJS",
            sofi_img_jitter_config.nbobjs);
    cpl_propertylist_append_double(qclist, "ESO QC IQ",
            sofi_img_jitter_config.iq);
    cpl_propertylist_append_double(qclist, "ESO QC FWHM PIX",
            sofi_img_jitter_config.fwhm_pix);
    cpl_propertylist_append_double(qclist, "ESO QC FWHM ARCSEC",
            sofi_img_jitter_config.fwhm_arcsec);
    cpl_propertylist_append_double(qclist, "ESO QC FWHM MODE",
            sofi_img_jitter_config.fwhm_mode);
    cpl_propertylist_append_int(qclist, "ESO QC NB_OBJ_F",
            sofi_img_jitter_config.nb_obj_frames);
    cpl_propertylist_append_int(qclist, "ESO QC NB_SKY_F",
            sofi_img_jitter_config.nb_sky_frames);
    cpl_propertylist_append_int(qclist, "ESO QC NB_REJ_F",
            sofi_img_jitter_config.nb_rej_frames);

    /* Write the image */
    irplib_dfs_save_image(set,
            parlist,
            set,
            combined,
            CPL_BPP_IEEE_FLOAT,
            "sofi_img_jitter",
            SOFI_IMG_JITTER_COMB,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            "sofi_img_jitter.fits");


    /* Write the FITS background table */
    if (sky_bg) {
        /* First create the table from the vector */
        sky_bg_tab = cpl_table_new(cpl_vector_get_size(sky_bg));
        cpl_table_new_column(sky_bg_tab, "SKY_BG", CPL_TYPE_DOUBLE);
        for (i=0 ; i<cpl_vector_get_size(sky_bg) ; i++) {
            cpl_table_set_double(sky_bg_tab, "SKY_BG", i,
                    cpl_vector_get(sky_bg, i));
        }
        irplib_dfs_save_table(set,
                parlist,
                set,
                sky_bg_tab,
                NULL,
                "sofi_img_jitter",
                SOFI_IMG_JITTER_BG,
                qclist,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                "sofi_img_jitter_bg.fits");
        cpl_table_delete(sky_bg_tab);
    }

    /* Write the FITS table */
    if (objs_stats) {
        irplib_dfs_save_table(set,
                parlist,
                set,
                objs_stats,
                NULL,
                "sofi_img_jitter",
                SOFI_IMG_JITTER_STARS,
                qclist,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                "sofi_img_jitter_stars.fits");
    }

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);

    /* Get FITS header from reference file */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist,
        "^(ARCFILE|MJD-OBS|INSTRUME|ESO TPL ID|ESO TPL NEXP|ESO DPR CATG|"
        "ESO DPR TECH|ESO DPR TYPE|DATE-OBS|ESO OBS ID|ESO INS PIXSCALE)$", 0);
    cpl_propertylist_delete(plist);

    /* Copy the QC in paflist */
    cpl_propertylist_copy_property_regexp(paflist, qclist, ".", 0);
    cpl_propertylist_delete(qclist);

    /* Save the PAF file */
    cpl_dfs_save_paf("SOFI",
            "sofi_img_jitter",
            paflist,
            "sofi_img_jitter.paf");
    cpl_propertylist_delete(paflist);
    return  0;
}
 
