/* $Id: sofi_spc_flat.c,v 1.32 2013-03-12 08:04:51 llundin Exp $
 *
 * This file is part of the SOFI Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:04:51 $
 * $Revision: 1.32 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>
#include <cpl.h>

#include "irplib_plugin.h"
#include "irplib_utils.h"

#include "sofi_utils.h"
#include "sofi_pfits.h"
#include "sofi_dfs.h"

/*-----------------------------------------------------------------------------
                            Defines
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "sofi_spc_flat"

#define MEDIAN_XSIZE        200
#define MEDIAN_YSIZE        200

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static cpl_image * sofi_spc_flat_reduce(cpl_frameset *);
static cpl_imagelist * sofi_spc_flat_diffs(cpl_frameset *);
static cpl_image * sofi_spc_flat_divide_fit(cpl_image *, int, int, int);
static cpl_error_code sofi_spc_flat_save(cpl_frameset *, int, const cpl_image *,
                                         const cpl_frameset *,
                                         const cpl_parameterlist *);
static int sofi_spc_flat_compare(const cpl_frame *, const cpl_frame *);

cpl_recipe_define(sofi_spc_flat, SOFI_BINARY_VERSION,
                  "Lars Lundin", PACKAGE_BUGREPORT, "2002,2003,2009", 
                  "SOFI spectro flat-field creation",
                  RECIPE_STRING " -- SOFI spectro flat-field creation.\n"
                  "The files listed in the Set Of Frames (sof-file) " 
                  "must be tagged:\n"
                  "raw-file.fits "SOFI_SPC_FLAT_RAW"\n");

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    double      low_thresh;
    double      high_thresh;
    int         fit_order;
    int         fit_size;
    int         offset;
    int         llx;
    int         lly;
    int         urx;
    int         ury;
    /* Outputs */
    double      med_stdev;
    double      med_avg;
} sofi_spc_flat_config;


/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Fill the recipe parameterlist
  @param    self  The parameterlist
  @return   0 iff everything is ok

 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code sofi_spc_flat_fill_parameterlist(cpl_parameterlist * self)
{
    const char * context = PACKAGE "." RECIPE_STRING;
    cpl_error_code err;

    cpl_ensure_code(self, CPL_ERROR_NULL_INPUT);

    /* Fill the parameters list */

    /* --thresholds */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING,
                                          "thresholds", "0.01,3.0", NULL,
                                          context, "Low and high thresholds");
    cpl_ensure_code(!err, err);

    /* --fit_order */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING, "fit_order",
                                       3, NULL, context, "Order of the fit");
    cpl_ensure_code(!err, err);

    /* --fit_size */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING, "fit_size",
                                       200, NULL, context, "Size of the fit");
    cpl_ensure_code(!err, err);

    /* --offset */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING, "offset",
                                       40, NULL, context, "Offset");
    cpl_ensure_code(!err, err);

    /* --zone */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING,
                                          "zone", "256,256,768,768", NULL,
                                          context, "Zone to process [pixel]");
    cpl_ensure_code(!err, err);

    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    The recipe data reduction part is implemented here 
  @param    framelist   the frames list
  @param    parlist     the parameters list
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static int sofi_spc_flat(cpl_frameset            * framelist,
                        const cpl_parameterlist * parlist)
{
    const int      nframes = cpl_frameset_get_size(framelist);
    const char   * sval;
    cpl_frameset * flatframes = NULL;
    cpl_size     * labels = NULL;
    cpl_size       nlabels;
    int            nflats;
    int            i;
    
    /* Initialise */
    sofi_spc_flat_config.med_stdev = 0.0;
    sofi_spc_flat_config.med_avg = 0.0;

    /* Retrieve input parameters */
    /* --thresholds */
    sval = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                       "thresholds");
    bug_if(sval == NULL);
    skip_if (sscanf(sval, "%lg,%lg",
                    &sofi_spc_flat_config.low_thresh,
                    &sofi_spc_flat_config.high_thresh) != 2);

    /* --fit_order */
    sofi_spc_flat_config.fit_order
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "fit_order");
    /* --fit_size */
    sofi_spc_flat_config.fit_size
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "fit_size");
    /* --offset */
    sofi_spc_flat_config.offset
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "offset");
    /* --zone */
    sval = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                           "zone");
    bug_if(sval == NULL);
    skip_if (sscanf(sval, "%d,%d,%d,%d",
                    &sofi_spc_flat_config.llx,
                    &sofi_spc_flat_config.lly,
                    &sofi_spc_flat_config.urx,
                    &sofi_spc_flat_config.ury) != 4);


    /* Identify the RAW and CALIB frames in the input frameset */
    skip_if (sofi_dfs_set_groups(framelist));

    /* Retrieve raw frames */
    flatframes = sofi_extract_frameset(framelist, SOFI_SPC_FLAT_RAW);
    error_if (flatframes == NULL, CPL_ERROR_DATA_NOT_FOUND,
              "Cannot find flat frames in the input frame list");
   
    nflats = cpl_frameset_get_size(flatframes);

    /* The number of frames must be even */
    error_if (cpl_frameset_get_size(flatframes) % 2, CPL_ERROR_ILLEGAL_INPUT,
              "The number of flat frames must be even, not %d", nflats);
    
    /* Labelise all input frames */
    labels = cpl_frameset_labelise(flatframes, sofi_spc_flat_compare, &nlabels);
    error_if (labels == NULL, cpl_error_get_code(),
              "Cannot labelise input frames");
   
    /* Extract settings and reduce each of them */
    for (i=0; i < nlabels; i++) {
        /* Reduce data set nb i */
        cpl_errorstate prestate = cpl_errorstate_get();
        cpl_frameset * flat_one = cpl_frameset_extract(flatframes, labels, i);
        cpl_image    * spflat;

        cpl_msg_info(cpl_func, "Reducing set %d/%d", i+1, (int)nlabels);

        spflat = sofi_spc_flat_reduce(flat_one);

        if (spflat == NULL) {
            cpl_msg_warning(cpl_func, "Could not reduce set %d:", i+1);
            cpl_errorstate_dump(prestate, CPL_FALSE,
                                irplib_errorstate_dump_warning);
            cpl_errorstate_set(prestate);
        } else {
            sofi_spc_flat_save(framelist, i+1, spflat, flat_one, parlist);
            cpl_image_delete(spflat);
        }
        cpl_frameset_delete(flat_one);
        if (!cpl_errorstate_is_equal(prestate)) break;
    }
    skip_if(i != nlabels);

    error_if (cpl_frameset_get_size(framelist) == nframes,
              CPL_ERROR_ILLEGAL_INPUT, "No products produced from %d flat "
              "frames grouped into %d set(s)", nflats, (int)nlabels);

    end_skip;
    
    cpl_frameset_delete(flatframes);
    cpl_free(labels);
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    flatframes  the frames list with flat frames
  @param    set_nb      the setting number
  @return   0 if ok, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_image * sofi_spc_flat_reduce(cpl_frameset * flatframes)
{
    cpl_imagelist       *   diffs;
    cpl_vector          *   medians;
    double                  med, mean;
    cpl_image           *   avg_ima;
    cpl_image           *   fitted;
    int                     nima, nx, ny;
    int                     nframes;
    int                     i;

    /* Test entries */
    if (flatframes == NULL) return NULL;
    nframes = cpl_frameset_get_size(flatframes);
    if (nframes != 4) {
        cpl_msg_error(cpl_func, "Expect 4 frames in input, not %d", nframes);
        return NULL;
    }
    
    /* Load input image set */
    cpl_msg_info(cpl_func, "Compute the difference images");
    if ((diffs = sofi_spc_flat_diffs(flatframes)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot create the difference images");
        return NULL;
    }
    nima = cpl_imagelist_get_size(diffs);
    nx = cpl_image_get_size_x(cpl_imagelist_get(diffs, 0));
    ny = cpl_image_get_size_y(cpl_imagelist_get(diffs, 0));

    /* Compute medians on diff images */
    cpl_msg_info(cpl_func, "Compute statistics on images");
    cpl_msg_indent_more();
    medians = cpl_vector_new(nima);

    /* Compute some stats on input images */
    for (i=0; i<nima; i++) {
        med = cpl_image_get_median_window(cpl_imagelist_get(diffs, i),
                (nx-MEDIAN_XSIZE)/2.0, (ny-MEDIAN_YSIZE)/2.0,
                (nx+MEDIAN_XSIZE)/2.0, (ny+MEDIAN_YSIZE)/2.0);
        if (cpl_vector_set(medians, i, med) != CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "Cannot compute the medians");
            cpl_vector_delete(medians);
            cpl_imagelist_delete(diffs);
            cpl_msg_indent_less();
            return NULL;
        }
    }

    /* Compute stdev and mean of the medians */
    sofi_spc_flat_config.med_avg   = cpl_vector_get_mean(medians);
    if (nima >= 2) 
        sofi_spc_flat_config.med_stdev=cpl_vector_get_stdev(medians);
    cpl_vector_delete(medians);
    cpl_msg_info(cpl_func, "Average of the medians: %g", 
            sofi_spc_flat_config.med_avg);
    cpl_msg_info(cpl_func, "Standard deviation of the medians: %g", 
            sofi_spc_flat_config.med_stdev);
    cpl_msg_indent_less();

    /* Divide by the mean in the vig in the difference image */
    cpl_msg_info(cpl_func, "Normalise the images");
    for (i=0; i<nima; i++) {
        mean = cpl_image_get_mean_window(cpl_imagelist_get(diffs, i),
                sofi_spc_flat_config.llx, sofi_spc_flat_config.lly,
                sofi_spc_flat_config.urx, sofi_spc_flat_config.ury);
        if (cpl_image_divide_scalar(cpl_imagelist_get(diffs, i), mean) != 
                CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "Cannot normalise the image");
            cpl_imagelist_delete(diffs);
            return NULL;
        }
    }

    /* Replace by 0 the pixels whose value is <low and >high */
    for (i=0; i<nima; i++) {
        if (cpl_image_threshold(cpl_imagelist_get(diffs, i), 
                sofi_spc_flat_config.low_thresh,
                sofi_spc_flat_config.high_thresh,
                0.0, 0.0) != CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "Cannot threshold the image");
            cpl_imagelist_delete(diffs);
            return NULL;
        }
    }

    /* Average the image list to one image */
    cpl_msg_info(cpl_func, "Stack the images");
    if ((avg_ima = cpl_imagelist_collapse_create(diffs)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot collapse the image list");
        cpl_imagelist_delete(diffs);
        return NULL;
    }
    cpl_imagelist_delete(diffs);

    /* Divide the output image by the fit    */
    cpl_msg_info(cpl_func, "Divide by a fit");
    if ((fitted = sofi_spc_flat_divide_fit(avg_ima,
                    sofi_spc_flat_config.fit_order,
                    sofi_spc_flat_config.fit_size,
                    sofi_spc_flat_config.offset)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot collapse the image list");
        cpl_image_delete(avg_ima);
        return NULL;
    }
    cpl_image_delete(avg_ima);
    
    return fitted;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create the difference images
  @param    in      the input frameset
  @return   the imagelist with the difference images
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * sofi_spc_flat_diffs(cpl_frameset * in) 
{
    int                     nima;
    cpl_imagelist       *   out;
    cpl_image           *   in1;
    cpl_image           *   in2;
    cpl_frame           *   cur_frame;
    int                     i;
 
    /* Initialise */
    nima = cpl_frameset_get_size(in);

    /* Expect an even number of frames */
    if ((nima % 2) || (nima != 4)) {
        cpl_msg_error(cpl_func, "Wrong nb of frames");
        return NULL;
    }

    /* Create the output image list */
    out = cpl_imagelist_new();

    /* Loop on all pairs */
    for (i=0; i<nima/2; i++) {
        cur_frame = cpl_frameset_get_position(in, 2*i);
        if ((in1 = cpl_image_load(cpl_frame_get_filename(cur_frame), 
                        CPL_TYPE_FLOAT, 0, 0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the image %d", 2*i);
            cpl_imagelist_delete(out);
            return NULL;
        }
        cur_frame = cpl_frameset_get_position(in, 2*i+1);
        if ((in2 = cpl_image_load(cpl_frame_get_filename(cur_frame), 
                        CPL_TYPE_FLOAT, 0, 0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the image %d", 2*i+1);
            cpl_image_delete(in1);
            cpl_imagelist_delete(out);
            return NULL;
        }

        /* Expected series: off on on off */
        if (i==0) {
            cpl_image_subtract(in2, in1);
            cpl_image_delete(in1);
            cpl_imagelist_set(out, in2, i);
        } else {
            cpl_image_subtract(in1, in2);
            cpl_image_delete(in2);
            cpl_imagelist_set(out, in1, i);
        }
    }
    return out;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Divide an image by a fit of itself
  @param    in      the input image
  @return   the result image
 */
/*----------------------------------------------------------------------------*/
static cpl_image * sofi_spc_flat_divide_fit(
        cpl_image   *   in,
        int             order,
        int             xsize,
        int             offset) 
{
    const cpl_size          degree = order - 1;
    int                     nx, ny;
    int                     xstart, xend, ystart, yend;
    cpl_image           *   collapsed;
    float               *   pcollapsed;
    cpl_image           *   extracted;
    float               *   pextracted;
    int                     nb_samples;
    cpl_matrix          *   xvec;
    cpl_vector          *   yvec;
    cpl_polynomial      *   poly_fit;
    cpl_polynomial      *   poly_2d;
    cpl_image           *   fit_image;
    cpl_image           *   out;
    cpl_size                power[2];
    int                     i;
    
    /* Test entries */
    if (in == NULL) return NULL;

    /* Initialise */
    nx = cpl_image_get_size_x(in);
    ny = cpl_image_get_size_y(in);
    
    /* Determine the zone to extract */
    xstart = (int)((nx - xsize)/2) + 1;
    xend = xstart + xsize - 1;
    if ((xstart<1) || (xend>nx)) {
        cpl_msg_error(cpl_func, "bad X size specified");
        return NULL;
    }

    /* Extract the central zone and collapse it */
    if ((collapsed = cpl_image_collapse_window_create(in, xstart, 1, xend, ny, 
                    1)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot collpase a part of the image");
        return NULL;
    }
    pcollapsed = cpl_image_get_data_float(collapsed);

    /* Find the 'valid' zone in the 1D image */
    ystart = 1;
    while ((fabs(pcollapsed[ystart-1]) < 1e-4) && (ystart < nx)) ystart++;
    ystart += offset;

    yend = ny;
    while ((fabs(pcollapsed[yend-1]) <1e-4) && (yend > 1)) yend--;
    yend -= offset;

    if (ystart > yend) {
        cpl_msg_error(cpl_func, "invalid coordinates of the zone to extract");
        cpl_image_delete(collapsed);
        return NULL;
    }

    /* Extract the 1D signal to fit */
    if ((extracted = cpl_image_extract(collapsed, 1, ystart, 1, yend))==NULL) {
        cpl_msg_error(cpl_func, "cannot extract 1D image");
        cpl_image_delete(collapsed);
        return NULL;
    }
    cpl_image_delete(collapsed);
    pextracted = cpl_image_get_data_float(extracted);

    /* Fit the polynomial */
    nb_samples = cpl_image_get_size_y(extracted);
    xvec = cpl_matrix_new(1, nb_samples);
    yvec = cpl_vector_new(nb_samples);
    for (i=0; i<nb_samples; i++) {
        cpl_matrix_set(xvec, 0, i, (double)(ystart + i));
        cpl_vector_set(yvec, i, (double)(pextracted[i] / xsize));
    }
    cpl_image_delete(extracted);
    poly_fit = cpl_polynomial_new(1);
    if (cpl_polynomial_fit(poly_fit, xvec, NULL, yvec, NULL, CPL_FALSE,
                           NULL, &degree)) {
        cpl_msg_error(cpl_func, "cannot fit the 1D signal");
        cpl_polynomial_delete(poly_fit);
        cpl_matrix_delete(xvec);
        cpl_vector_delete(yvec);
        return NULL;
    }
    cpl_matrix_delete(xvec);
    cpl_vector_delete(yvec);
    
    /* The polynomial for image generation must be 2d */
    poly_2d = cpl_polynomial_new(2);
    power[0] = 0; power[1] = 0;
    cpl_polynomial_set_coeff(poly_2d, power, 
            cpl_polynomial_get_coeff(poly_fit, power + 1));
    power[0] = 0; power[1] = 1;
    cpl_polynomial_set_coeff(poly_2d, power, 
            cpl_polynomial_get_coeff(poly_fit, power + 1));
    power[0] = 0; power[1] = 2;
    cpl_polynomial_set_coeff(poly_2d, power, 
            cpl_polynomial_get_coeff(poly_fit, power + 1));
    cpl_polynomial_delete(poly_fit);
    
    /* Create the fit image */
    fit_image = cpl_image_new(nx, ny, CPL_TYPE_FLOAT);
    cpl_image_fill_polynomial(fit_image, poly_2d, 1.0, 1.0, 1.0, 1.0);
    cpl_polynomial_delete(poly_2d);
    
    /* Divide the input image by the polynomial image */
    if ((out = cpl_image_divide_create(in, fit_image)) == NULL) {
        cpl_msg_error(cpl_func, "cannot divide the images");
        cpl_image_delete(fit_image);
        return NULL;
    }
    cpl_image_delete(fit_image);

    return out;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the lampflat recipe products on disk
  @param    set_tot the input frame set
  @param    set_nb  current setting number
  @param    flat    the flat image produced
  @param    set     the frames from the current setting
  @param    parlist the input list of parameters
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code sofi_spc_flat_save(cpl_frameset            * set_tot,
                                         int                       set_nb,
                                         const cpl_image         * flat,
                                         const cpl_frameset      * set,
                                         const cpl_parameterlist * parlist)
{
#define SOFI_SPC_FLAT_PAF                                               \
    "^(ARCFILE|MJD-OBS|INSTRUME|ESO TPL ID|ESO TPL NEXP|ESO DPR CATG"   \
        "|ESO DPR TECH|ESO DPR TYPE|DATE-OBS|ESO INS GRAT NAME"         \
        "|ESO INS GRAT WLEN|ESO INS OPTI1 ID|ESO DET DIT|ESO INS LAMP3 SET)$"

    cpl_errorstate        prestate = cpl_errorstate_get();
    /* Get the reference frame */
    const cpl_frame     * ref_frame
        = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);
    /* Get the QC params in qclist */
    cpl_propertylist    * qclist = cpl_propertylist_new();
    cpl_propertylist    * plist  = cpl_propertylist_load_regexp
        (cpl_frame_get_filename(ref_frame), 0, SOFI_SPC_FLAT_PAF
         "|^ESO INS FILT[12] ID$", 0);
    const char          * sval;
    char                * filename;

    skip_if(0);

    sval = sofi_pfits_get_filter(plist);
    if (sval == NULL) {
        cpl_msg_warning(cpl_func, "Could not get filter for set %d:", set_nb);
        cpl_errorstate_dump(prestate, CPL_FALSE,
                            irplib_errorstate_dump_warning);
        cpl_errorstate_set(prestate);
    } else {
        cpl_propertylist_append_string(qclist, "ESO QC FILTER OBS", sval);
    }

    cpl_propertylist_append_double(qclist, "ESO QC SPECFLAT NCOUNTS",
            sofi_spc_flat_config.med_avg);
    cpl_propertylist_append_double(qclist, "ESO QC SPECFLAT STDEV",
            sofi_spc_flat_config.med_stdev);

    /* Write the flat image */
    filename = cpl_sprintf(RECIPE_STRING "_set%02d" CPL_DFS_FITS, set_nb);
    irplib_dfs_save_image(set_tot, parlist, set, flat, CPL_BPP_IEEE_FLOAT,
                          RECIPE_STRING, SOFI_SPC_FLAT_RES, qclist, NULL,
                          PACKAGE "/" PACKAGE_VERSION, filename);
    cpl_free(filename);
    skip_if(0);

    cpl_propertylist_copy_property_regexp(qclist, plist,
                                          SOFI_SPC_FLAT_PAF, 0);
    cpl_propertylist_empty(plist);

    /* Save the PAF file */
    filename = cpl_sprintf(RECIPE_STRING "_set%02d" CPL_DFS_PAF, set_nb);
    cpl_dfs_save_paf("SOFI", RECIPE_STRING, qclist, filename);
    cpl_free(filename);

    skip_if(0);

    end_skip;

    cpl_propertylist_delete(qclist);
    cpl_propertylist_delete(plist);

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Comparison function to identify different settings
  @param    frame1  first frame 
  @param    frame2  second frame 
  @return   0 if frame1!=frame2, 1 if frame1==frame2, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static int sofi_spc_flat_compare(
        const cpl_frame *   frame1, 
        const cpl_frame *   frame2) 
{
    int                     comparison;
    cpl_propertylist    *   plist1;
    cpl_propertylist    *   plist2;
    const char          *   sval1,
                        *   sval2;

    /* Test entries */
    if (frame1==NULL || frame2==NULL) return CPL_ERROR_UNSPECIFIED;

    /* Get property lists */
    if ((plist1=cpl_propertylist_load(cpl_frame_get_filename(frame1),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        return CPL_ERROR_UNSPECIFIED;
    }
    if ((plist2=cpl_propertylist_load(cpl_frame_get_filename(frame2),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(plist1);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Test status */
    if (cpl_error_get_code()) {
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }

    comparison = 1;

    /* Compare the slit used */
    sval1 = sofi_pfits_get_opti1_id(plist1);
    sval2 = sofi_pfits_get_opti1_id(plist2);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get the slit used");
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }
    if (strcmp(sval1, sval2)) comparison = 0;

    /* Compare the filter */
    sval1 = sofi_pfits_get_filter(plist1);
    sval2 = sofi_pfits_get_filter(plist2);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get the filter");
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }
    if (strcmp(sval1, sval2)) comparison = 0;

    cpl_propertylist_delete(plist1);
    cpl_propertylist_delete(plist2);
    return comparison;
}


