/* $Id: sofi_spc_jitter.c,v 1.45 2013-03-12 08:04:50 llundin Exp $
 *
 * This file is part of the SOFI Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:04:50 $
 * $Revision: 1.45 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>
#include <cpl.h>

#include "irplib_plugin.h"
#include "irplib_utils.h"
#include "irplib_spectrum.h"

#include "sofi_utils.h"
#include "sofi_wavelength.h"
#include "sofi_pfits.h"
#include "sofi_dfs.h"

/*-----------------------------------------------------------------------------
                            Defines
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "sofi_spc_jitter"

#define SOFI_SPC_JITTER_OFFSET_ERR     10

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static cpl_image ** sofi_spc_jitter_combine(cpl_frameset *, const char *, 
        const char *, const char *);
static cpl_vector * sofi_spc_jitter_get_offsets(cpl_frameset *);
static int * sofi_spc_jitter_classif(cpl_vector *, int *);
static int off_comp(double, double, double);
static cpl_imagelist * sofi_spc_jitter_saa_groups(cpl_imagelist *,
        cpl_vector *, int *, int, cpl_vector **);
static int sofi_spc_jitter_wavecal(const char *, const char *, cpl_image *, 
        cpl_frameset *);
static cpl_imagelist * sofi_spc_jitter_nodded(cpl_imagelist *, cpl_vector *, 
        cpl_vector **);
static cpl_imagelist * sofi_spc_jitter_distor(cpl_imagelist *, const char *);
static double sofi_spc_jitter_refine_offset(cpl_image *, cpl_image *);
static cpl_table * sofi_spc_jitter_extract(cpl_image *);
static cpl_error_code sofi_spc_jitter_save(cpl_frameset *, const cpl_image *,
                                           const cpl_table *,
                                           const cpl_parameterlist *);

cpl_recipe_define(sofi_spc_jitter, SOFI_BINARY_VERSION,
                  "Lars Lundin", PACKAGE_BUGREPORT, "2002,2003,2009", 
                  "SOFI Spectro jitter recipe",
                  RECIPE_STRING " -- SOFI Spectro jitter recipe"
                  "The files listed in the Set Of Frames (sof-file) " 
                  "must be tagged:\n"
                  "raw-file.fits "SOFI_SPC_JITTER_NODOBJ_RAW" or\n"
                  "raw-file.fits "SOFI_SPC_JITTER_NODSKY_RAW"\n"
                  "Calibration files:\n"
                  "oh-cat.fits "SOFI_CALPRO_OH_CAT" or\n"
                  "flat-file.fits "SOFI_CALIB_SPFLAT" or\n"
                  "arc-file.fits "SOFI_CALIB_ARC"\n");

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int             display;
    int             crosstalk;
    /* wavecal_in : 0 for physical model, 1 for sky, 2 for arc table */
    int             wavecal_in;
    int             wavecal_degree;
    double          wavecal_err;
    int             wavecal_nsamples;
    int             wavecal_ppm;
    int             saa_refine;
    double          saa_rej_high;
    double          saa_rej_low;
    int             extr_spec_pos;
    int             extr_spec_width;
    int             extr_sky_le_width;
    int             extr_sky_ri_width;
    int             extr_sky_le_dist;
    int             extr_sky_ri_dist;
    /* Outputs */
    /* wavecal_out : 0 for physical model, 1 for sky, 2 for arc table */
    cpl_vector  *   throws;
    cpl_vector  *   intensities;
    char            filter[512];
    int             wavecal_out;
    double          wavecal_cc;
    double          wavecal_a0;
    double          wavecal_a1;
    double          wavecal_a2;
    double          wavecal_a3;
    double          wavecal_a4;
} sofi_spc_jitter_config;


/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Fill the recipe parameterlist
  @param    self  The parameterlist
  @return   0 iff everything is ok

 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code sofi_spc_jitter_fill_parameterlist(cpl_parameterlist * self)
{
    const char * context = PACKAGE "." RECIPE_STRING;
    cpl_error_code err;

    cpl_ensure_code(self, CPL_ERROR_NULL_INPUT);

    /* Fill the parameters list */


    /* --crosstalk */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING,
                                        "crosstalk", CPL_TRUE, NULL, context,
                                        "Enable removal of the crosstalk "
                                        "effect");
    cpl_ensure_code(!err, err);

    /* --wavecal */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING,
                                          "wavecal", "sky", NULL, context,
                                          "Wavelength method: "
                                          "phy or sky or arc");
    cpl_ensure_code(!err, err);

    /* --wavecal_ppm */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING,
                                        "wavecal_ppm", CPL_FALSE, NULL, context,
                                        "Enable Point Pattern Matching in the "
                                        "wavelength calibration");
    cpl_ensure_code(!err, err);

    /* --wavecal_degree */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "wavecal_degree", 2, NULL, context,
                                       "Degree of the wavelength dispersion "
                                       "polynomial");
    cpl_ensure_code(!err, err);

    /* --wavecal_nsamples */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "wavecal_nsamples", 100, NULL, context,
                                       "Number of samples for the wavelength "
                                       "calibration");
    cpl_ensure_code(!err, err);

    /* --wavecal_err */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING,
                                          "wavecal_err", 1000.0, NULL, context,
                                          "The wavelength error [Angstrom]");
    cpl_ensure_code(!err, err);

    /* --saa_refine */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING,
                                        "saa_refine", CPL_TRUE, NULL, context,
                                        "Enable refinement of the offsets");
    cpl_ensure_code(!err, err);

    /* --saa_rej */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING,
                                          "saa_rej", "0.1,0.1", NULL, context,
                                          "Low, high SAA rejection [%,%]");
    cpl_ensure_code(!err, err);

    /* --spec_pos */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "spec_pos", -1, NULL, context,
                                       "Spectrum position");
    cpl_ensure_code(!err, err);

    /* --spec_width */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "spec_width", 10, NULL, context,
                                       "Spectrum width");
    cpl_ensure_code(!err, err);

    /* --sky_le_width */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "sky_le_width", 10, NULL, context,
                                       "Sky width left of the spectrum");
    cpl_ensure_code(!err, err);

    /* --sky_ri_width */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "sky_ri_width", 10, NULL, context,
                                       "Sky width right of the spectrum");
    cpl_ensure_code(!err, err);

    /* --sky_le_dist */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "sky_le_dist", -1, NULL, context,
                                       "Sky distance left of the spectrum");
    cpl_ensure_code(!err, err);

    /* --sky_ri_dist */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "sky_ri_dist", -1, NULL, context,
                                       "Sky distance right of the spectrum");
    cpl_ensure_code(!err, err);

    /* --display */ 
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING, "display",
                                        CPL_FALSE, NULL, context,
                                        "Enable plotting");
    cpl_ensure_code(!err, err);

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int sofi_spc_jitter(cpl_frameset            * framelist,
                           const cpl_parameterlist * parlist)
{
    cpl_propertylist    *   plist;
    const char          *   sval;
    cpl_frameset        *   rawframes = NULL;
    const cpl_frame     *   cur_frame;
    const char          *   flat;
    const char          *   oh;
    const char          *   arc;
    cpl_image           **  combined = NULL;
    cpl_table           *   extracted = NULL;
    
    /* Initialise */
    sofi_spc_jitter_config.wavecal_out = -1;
    sofi_spc_jitter_config.wavecal_cc = -1.0;
    sofi_spc_jitter_config.throws = NULL;
    sofi_spc_jitter_config.intensities = NULL;
    sofi_spc_jitter_config.filter[0] = (char)0;

    /* Retrieve input parameters */

    /* Cross-talk */
    sofi_spc_jitter_config.crosstalk
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "crosstalk");

    /* --wavecal */
    sval = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                           "wavecal");
    bug_if(sval == NULL);
    if (!strcmp(sval, "phy")) sofi_spc_jitter_config.wavecal_in = 0;
    else if (!strcmp(sval, "sky")) sofi_spc_jitter_config.wavecal_in = 1;
    else if (!strcmp(sval, "arc")) sofi_spc_jitter_config.wavecal_in = 2;
    else {
        error_if(1, CPL_ERROR_UNSUPPORTED_MODE,
                 "Invalid value for wavecal option: %s", sval);
    }

    /* --wavecal_degree */
    sofi_spc_jitter_config.wavecal_degree
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "wavecal_degree");

    /* --wavecal_nsamples */
    sofi_spc_jitter_config.wavecal_nsamples
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "wavecal_nsamples");

    /* --wavecal_err */
    sofi_spc_jitter_config.wavecal_err
        = irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING,
                                          "wavecal_err");

    /* --wavecal_ppm */
    sofi_spc_jitter_config.wavecal_ppm
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "wavecal_ppm");

    /* Refine of offsets */
    sofi_spc_jitter_config.saa_refine
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "saa_refine");
    
    /* Rejection parameters for SAA */
    sval = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                           "saa_rej");
    skip_if (sscanf(sval, "%lg,%lg",
                    &sofi_spc_jitter_config.saa_rej_low,
                    &sofi_spc_jitter_config.saa_rej_high) != 2);

    /* --spec_pos */
    sofi_spc_jitter_config.extr_spec_pos
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "spec_pos");

    /* --spec_width */
    sofi_spc_jitter_config.extr_spec_width
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "spec_width");

    /* --sky_le_width */
    sofi_spc_jitter_config.extr_sky_le_width
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "sky_le_width");
    /* --sky_ri_width */
    sofi_spc_jitter_config.extr_sky_ri_width
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "sky_ri_width");

    /* --sky_le_dist */
    sofi_spc_jitter_config.extr_sky_le_dist
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "sky_le_dist");

    /* --sky_ri_dist */
    sofi_spc_jitter_config.extr_sky_ri_dist
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "sky_ri_dist");
    /* Display */
    sofi_spc_jitter_config.display
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "display");
    
    /* Identify the RAW and CALIB frames in the input frameset */
    skip_if (sofi_dfs_set_groups(framelist));

    /* Get the filter */
    cur_frame = cpl_frameset_get_position(framelist, 0);
    plist = cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
    skip_if(plist == NULL);

    if ((sval = sofi_pfits_get_filter(plist)) != NULL) {
        strcpy(sofi_spc_jitter_config.filter, sval); /* FIXME: Replace */
    } else {
        cpl_error_reset();
    }
    cpl_propertylist_delete(plist);

    /* Retrieve calibration data */
    arc =  sofi_extract_filename(framelist, SOFI_CALIB_ARC);
    flat = sofi_extract_filename(framelist, SOFI_CALIB_SPFLAT);
    oh = sofi_extract_filename(framelist, SOFI_CALPRO_OH_CAT);

    /* Retrieve raw frames */
    rawframes = sofi_extract_frameset(framelist, SOFI_SPC_JITTER_NODOBJ_RAW);
    error_if (rawframes == NULL, CPL_ERROR_DATA_NOT_FOUND,
              "Cannot find the raw frames in the input list");

    /* If arc calibration requested, arc file must be provided */
    error_if (arc == NULL && sofi_spc_jitter_config.wavecal_in == 2,
              CPL_ERROR_ILLEGAL_INPUT,
              "You must provide an ARC file for the wl cal");

    /* Create the combined image */
    cpl_msg_info(cpl_func, "Create the combined image");
    cpl_msg_indent_more();
    combined = sofi_spc_jitter_combine(rawframes, oh, flat, arc);
    cpl_msg_indent_less();

    error_if (combined==NULL, CPL_ERROR_ILLEGAL_INPUT,
              "Cannot combine the images");
    
    /* Extract the spectrum */
    cpl_msg_info(cpl_func, "Extract the spectrum");
    cpl_msg_indent_more();
    extracted = sofi_spc_jitter_extract(combined[0]);
    cpl_msg_indent_less();

    error_if (extracted == NULL, CPL_ERROR_DATA_NOT_FOUND,
              "Cannot extract the spectrum");

    /* Write the products */
    cpl_msg_info(cpl_func, "Save the products");
    skip_if(sofi_spc_jitter_save(framelist, combined[0], extracted, parlist));

    end_skip;

    if (combined != NULL) {
        cpl_image_delete(combined[0]);
        cpl_image_delete(combined[1]);
        cpl_free(combined);
    }

    cpl_vector_delete(sofi_spc_jitter_config.intensities);
    cpl_vector_delete(sofi_spc_jitter_config.throws);

    cpl_frameset_delete(rawframes);

    cpl_table_delete(extracted);
 
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The image combination routine
  @param    rawframes   the spectrum frames
  @param    oh          the oh lines table
  @param    flat        the flat frame or NULL
  @param    arc_dist    the arc frame for distortion or NULL
  @return   the combined image or NULL
 */
/*----------------------------------------------------------------------------*/
static cpl_image ** sofi_spc_jitter_combine(
        cpl_frameset        *   rawframes,
        const char          *   oh,
        const char          *   flat,
        const char          *   arc)
{
    cpl_imagelist       *   ilist;
    cpl_image           *   tmp_im;
    cpl_vector          *   offsets;
    int                 *   groups;
    int                     ngroups;
    cpl_imagelist       *   abba;
    cpl_vector          *   abba_off;
    cpl_imagelist       *   nodded;
    cpl_vector          *   nodded_off_x;
    cpl_vector          *   nodded_off_y;
    double                  throw;
    cpl_table           *   extracted;
    double                  intensity;
    double              *   pnodded_off_x;
    cpl_imagelist       *   nodded_warped;
    cpl_bivector        *   nodded_offsets;
    cpl_image           **  combined;
    int                     nima;
    double                  new_offset;
    int                     i;
    
    /* Test entries */
    if (rawframes == NULL) return NULL;

    /* Load the images */
    cpl_msg_info(cpl_func, "Load the data");
    cpl_msg_indent_more();
    if ((ilist = cpl_imagelist_load_frameset(rawframes, CPL_TYPE_FLOAT, 1, 
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "cannot load the data");
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_msg_indent_less();
  
    /* Apply the cross-talk correction */
    if (sofi_spc_jitter_config.crosstalk) {
        cpl_msg_info(cpl_func, "Apply the cross-talk correction");
        cpl_msg_indent_more();
        if (sofi_correct_crosstalk_list(ilist) == -1) {
            cpl_msg_error(cpl_func, "Cannot correct for Cross-talk");
            cpl_imagelist_delete(ilist);
            cpl_msg_indent_less();
            return NULL;
        }
        cpl_msg_indent_less();
    }
    
    /* Apply the flafield */
    if (flat != NULL) {
        cpl_msg_info(cpl_func, "Apply the flatfield correction");
        if ((tmp_im = cpl_image_load(flat, CPL_TYPE_FLOAT, 0, 0)) == NULL) {
            cpl_msg_warning(cpl_func, "cannot load the flat field");
        } else {
            if (cpl_imagelist_divide_image(ilist, tmp_im) != CPL_ERROR_NONE) {
                cpl_msg_warning(cpl_func, "cannot apply the flat field");
            }
            cpl_image_delete(tmp_im);
        }
    }

    /* Get the offsets */
    cpl_msg_info(cpl_func, "Get the offsets");
    if ((offsets = sofi_spc_jitter_get_offsets(rawframes)) == NULL) {
        cpl_msg_error(cpl_func, "cannot get the offsets");
        cpl_imagelist_delete(ilist);
        return NULL;
    }

    /* Classify in groups the a and b images sequence */
    cpl_msg_info(cpl_func, "Classify in groups");
    cpl_msg_indent_more();
    if ((groups = sofi_spc_jitter_classif(offsets, &ngroups)) == NULL) {
        cpl_msg_error(cpl_func, "cannot classify the data");
        cpl_imagelist_delete(ilist);
        cpl_vector_delete(offsets);
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_msg_indent_less();
  
    /* Shift and add each group to one image */
    cpl_msg_info(cpl_func, "Shift and add each group to one image");
    cpl_msg_indent_more();
    if ((abba = sofi_spc_jitter_saa_groups(ilist, offsets, groups, 
                    ngroups, &abba_off)) == NULL) {
        cpl_msg_error(cpl_func, "cannot shift and add groups");
        cpl_imagelist_delete(ilist);
        cpl_vector_delete(offsets);
        cpl_free(groups);
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_imagelist_delete(ilist);
    cpl_free(groups);
    cpl_vector_delete(offsets);
    cpl_msg_indent_less();

    /* Compute the wavelength calibration */
    cpl_msg_info(cpl_func, "Compute the wavelength calibration");
    cpl_msg_indent_more();
    if (sofi_spc_jitter_wavecal(arc, oh, cpl_imagelist_get(abba, 0),
                rawframes) == -1) {
        cpl_msg_error(cpl_func, "cannot compute the wavelength");
        cpl_imagelist_delete(abba);
        cpl_vector_delete(abba_off);
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_msg_indent_less();
   
   /* Create the nodded images */
    cpl_msg_info(cpl_func, "Create the nodded images");
    cpl_msg_indent_more();
    if ((nodded = sofi_spc_jitter_nodded(abba, abba_off, 
                    &nodded_off_x))==NULL) {
        cpl_msg_error(cpl_func, "cannot create the nodded images");
        cpl_imagelist_delete(abba);
        cpl_vector_delete(abba_off);
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_imagelist_delete(abba);
    cpl_msg_indent_less();

    /* Get the throw offsets from abba_off */
    nima = cpl_imagelist_get_size(nodded);
    sofi_spc_jitter_config.throws = cpl_vector_new(nima);
    for (i=0; i<nima/2; i++) {
        throw = fabs(   (cpl_vector_get(abba_off, 2*i))-
                        (cpl_vector_get(abba_off, 2*i+1)));
        cpl_vector_set(sofi_spc_jitter_config.throws, 2*i, throw);
        cpl_vector_set(sofi_spc_jitter_config.throws, 2*i+1, throw);
    }
    cpl_vector_delete(abba_off);
    
    /* Get the spectra intensities in the nodded images */
    cpl_msg_info(cpl_func, "Compute the spectra intensities");
    cpl_msg_indent_more();
    nima = cpl_imagelist_get_size(nodded);
    sofi_spc_jitter_config.intensities = cpl_vector_new(nima);
    for (i=0; i<nima; i++) {
        if ((extracted = sofi_spc_jitter_extract(
                        cpl_imagelist_get(nodded, i))) == NULL) {
            cpl_msg_warning(cpl_func,"Cannot extract the spectrum from nodded %d", 
                    i+1);
            intensity = -1.0;
        } else {
            intensity = cpl_table_get_column_mean(extracted, 
                    "Extracted_spectrum_value");
            intensity *= cpl_table_get_nrow(extracted);
            cpl_table_delete(extracted);
        }
        cpl_msg_info(cpl_func, "Spectrum intensity nb %d: %g", i+1,
                intensity);
        cpl_vector_set(sofi_spc_jitter_config.intensities, i, intensity);
    }
    cpl_msg_indent_less();

    /* Distortion correction */
    if (arc) {
        cpl_msg_info(cpl_func, "Correct the distortion on nodded images");
        cpl_msg_indent_more();
        if ((nodded_warped = sofi_spc_jitter_distor(nodded, arc)) == NULL) {
            cpl_msg_error(cpl_func, "cannot correct the distortion");
            cpl_imagelist_delete(nodded);
            cpl_vector_delete(nodded_off_x);
            cpl_msg_indent_less();
            return NULL;
        }
        cpl_imagelist_delete(nodded);
        nodded = nodded_warped;
        cpl_msg_indent_less();
    }

    /* Refine the offsets if requested */
    if (sofi_spc_jitter_config.saa_refine) {
        cpl_msg_info(cpl_func, "Refine the offsets");
        pnodded_off_x = cpl_vector_get_data(nodded_off_x);
        for (i=0; i<cpl_imagelist_get_size(nodded); i++) {
            new_offset = sofi_spc_jitter_refine_offset(
                    cpl_imagelist_get(nodded, 0),
                    cpl_imagelist_get(nodded, i));
            if (new_offset > 5000) {
                cpl_msg_debug(cpl_func, "cannot refine the offset - keep %g", 
                        pnodded_off_x[i]);
            } else {
                if (fabs(new_offset-pnodded_off_x[i]) < 
                        SOFI_SPC_JITTER_OFFSET_ERR) {
                    cpl_msg_debug(cpl_func, "refined offset : %g (old was %g)",
                            new_offset, pnodded_off_x[i]);
                    pnodded_off_x[i] = new_offset;
                } else { 
                    cpl_msg_debug(cpl_func, 
                            "refined offset %g too different - keep %g",
                            new_offset, pnodded_off_x[i]);
                }
            }
        }
    }

    /* Images combination */
    /* Get the offsets in a bivector */
    nodded_off_y = cpl_vector_duplicate(nodded_off_x);
    cpl_vector_fill(nodded_off_y, 0.0);
    nodded_offsets = cpl_bivector_wrap_vectors(nodded_off_x, nodded_off_y);
    /* Shift and add */
    cpl_msg_info(cpl_func, "Apply the shift and add on the nodded frames");
    nima = cpl_imagelist_get_size(nodded);
    if ((combined = cpl_geom_img_offset_saa(nodded, nodded_offsets,
                    CPL_KERNEL_DEFAULT, 
                    (int)(sofi_spc_jitter_config.saa_rej_low * nima), 
                    (int)(sofi_spc_jitter_config.saa_rej_high * nima), 
		    CPL_GEOM_FIRST, NULL, NULL)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot shift and add group");
        cpl_imagelist_delete(nodded);
        cpl_bivector_unwrap_vectors(nodded_offsets);
        cpl_vector_delete(nodded_off_x);
        cpl_vector_delete(nodded_off_y);
        return NULL;
    }
    cpl_imagelist_delete(nodded);
    cpl_bivector_unwrap_vectors(nodded_offsets);
    cpl_vector_delete(nodded_off_x);
    cpl_vector_delete(nodded_off_y);
    return combined;
}
 
/*----------------------------------------------------------------------------*/
/**
  @brief    Get the offsets
  @param    rawframes   the input rawframes
  @return   the offsets in a vector
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * sofi_spc_jitter_get_offsets(cpl_frameset * rawframes)
{
    cpl_vector          *   offsets;
    double              *   pvect;
    int                     nraw;
    cpl_frame           *   cur_frame;
    cpl_propertylist    *   plist;
    int                     i;

    /* Test entries */
    if (rawframes == NULL) return NULL;
    
    /* Initialise */
    nraw = cpl_frameset_get_size(rawframes);

    /* Get the rawframes X offsets */
    offsets = cpl_vector_new(nraw);
    pvect = cpl_vector_get_data(offsets);
    for (i=0; i<nraw; i++) {
        cur_frame = cpl_frameset_get_position(rawframes, i);
        if ((plist=cpl_propertylist_load(cpl_frame_get_filename(cur_frame),
                        0)) == NULL) {
            cpl_msg_error(cpl_func, "cannot get property list");
            cpl_vector_delete(offsets);
            return NULL;
        }
        pvect[i] = sofi_pfits_get_cumoffsetx(plist);
        if (cpl_error_get_code()) {
            cpl_msg_error(cpl_func, "cannot get the offset from the header");
            cpl_vector_delete(offsets);
            cpl_propertylist_delete(plist);
            return NULL;
        }
        cpl_propertylist_delete(plist);
        pvect[i] *= -1.0;
    }
    return offsets;
}
    
/*----------------------------------------------------------------------------*/
/**
  @brief    The abba classification
  @param    offsets     the images offsets
  @param    ngroups     the computed number of groups (must be even)
  @return   the groups array
 
  The classification is done according the offsets read in header of the 
  rawframes.
  
  The frames are first classified in two different categories A or B depending 
  in which zone their offset is.

  Example
  frame1 - offset: -250.4    ---> cat. A
  frame2 - offset: -120.6    ---> cat. B
  frame3 - offset: -150.6    ---> cat. B
  frame4 - offset: -220.4    ---> cat. A

  These frames are then grouped in cubes according the category order.
  Each group is then averaged in a single image.

  Example 
  frame1 -  cat A    -> group 1  
  frame2 -  cat A    -> group 1
  frame3 -  cat B    -> group 2 
  frame4 -  cat B    -> group 2
  frame5 -  cat B    -> group 3
  frame6 -  cat B    -> group 3
  frame7 -  cat A    -> group 4
  frame8 -  cat A    -> group 4
  frame9 -  cat A    -> group 5
  frame10 - cat A    -> group 5
  frame11 - cat B    -> group 6
  frame12 - cat B    -> group 6
  frame13 - cat B    -> group 6
  frame14 - cat B    -> group 6
 */
/*----------------------------------------------------------------------------*/
static int * sofi_spc_jitter_classif(
        cpl_vector      *   offsets,
        int             *   ngroups)
{
    double              *   pvect;
    int                     nraw;
    double                  offset_thresh;
    cpl_vector          *   tmp_vec;
    int                 *   groups;
    int                     last_group;
    int                     i, j, k, l;

    /* Test entries */
    if (offsets == NULL) return NULL;

    /* Initialise */
    nraw = cpl_vector_get_size(offsets);

    /* Separate the offsets in 2 categories */
    tmp_vec = cpl_vector_duplicate(offsets);
    cpl_vector_sort(tmp_vec, 1);
    pvect = cpl_vector_get_data(tmp_vec);
    if (pvect[0] == pvect[nraw-1]) {
        cpl_msg_error(cpl_func, "Only one offset in the list - abort");
        cpl_vector_delete(tmp_vec);
        return NULL;
    }
    offset_thresh = (pvect[0] + pvect[nraw-1]) / 2.0;
    cpl_vector_delete(tmp_vec);

    /* Identify the different A and B groups */
    pvect = cpl_vector_get_data(offsets);
    *ngroups = 0;
    groups = cpl_calloc(nraw, sizeof(int));

    /* Create a look up table to associate the ith obj with the jth frame */
    i = 0;
    while (i < nraw) {
        j = 0;
        /* Count the number of successive '+' or '-' (j) */
        while ((i+j<nraw) &&
                (!off_comp(pvect[i], pvect[i+j], offset_thresh))) j++;

        if (i+j >= nraw) i = nraw;
        else {
            k = 0;
            /* Check if there are j '-' or '+' (k) */
            while ((i+j+k < nraw)
                    && (!off_comp(pvect[i+j], pvect[i+j+k], offset_thresh))
                    && (k<j)) k++;
            last_group = 1;
            if (i+j+k < nraw) {
                for (l=i+j+k; l<nraw; l++) {
                    if (off_comp(pvect[i+j], pvect[l], offset_thresh)) {
                        last_group = 0;
                        break;
                    }
                }
            }
            if (last_group == 0) {
                for (l=0; l<j; l++) groups[i+l]   = *ngroups + 1;
                for (l=0; l<k; l++) groups[i+j+l] = *ngroups + 2;
                *ngroups += 2;
                i += j+k;
            } else {
                for (l=0; l<j; l++)               groups[i+l] = *ngroups + 1;
                for (l=0; l<nraw - (i+j); l++)    groups[i+j+l] =*ngroups + 2;
                *ngroups += 2;
                i = nraw;
            }
        }
    }

    /* Nb of groups found should be even */
    if (*ngroups % 2) {
        cpl_msg_error(cpl_func, "Odd number of groups found");
        cpl_free(groups);
        return NULL;
    }

    return groups;
}
    
/*----------------------------------------------------------------------------*/
/**
  @brief    Shift and add groups
  @param    ilist       the input image list
  @param    offsets     the offsets
  @param    groups      the groups ids
  @param    ngroups     the number of groups
  @param    abba_off    the output list offsets
  @return   the new abba image list
 
  The frames in each group are combined together. In principle, there
  should be no offset difference between the frames in the same group,
  so an average should be enough. But to support more general cases
  (generic offset) a shift and add is applied if the offsets differ.

  Example 
  frame1 -  cat A    -> group 1  ---> shift and add 1
  frame2 -  cat A    -> group 1
  frame3 -  cat B    -> group 2  ---> shift and add 2
  frame4 -  cat B    -> group 2
  frame5 -  cat B    -> group 3  ---> shift and add 3
  frame6 -  cat B    -> group 3
  frame7 -  cat A    -> group 4  ---> shift and add 4
  frame8 -  cat A    -> group 4
  frame9 -  cat A    -> group 5  ---> shift and add 5
  frame10 - cat A    -> group 5
  frame11 - cat B    -> group 6  ---> shift and add 6
  frame12 - cat B    -> group 6
  frame13 - cat B    -> group 6
  frame14 - cat B    -> group 6
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * sofi_spc_jitter_saa_groups(
        cpl_imagelist   *   ilist,
        cpl_vector      *   offsets,
        int             *   groups,
        int                 ngroups,
        cpl_vector      **  abba_off)
{
    cpl_imagelist       *   abba;
    cpl_imagelist       *   group_list;
    cpl_image           *   tmp_ima;
    cpl_image           **  combined;
    cpl_bivector        *   group_off;
    double              *   pgroup_off;
    double              *   poffsets;
    double              *   pabba_off;
    int                     nima;
    int                     saa;
    int                     i, j, k;

    /* Test entries */
    if ((ilist == NULL) || (offsets == NULL) || (groups == NULL)) return NULL;
    
    /* Initialise */
    nima = cpl_imagelist_get_size(ilist);
    poffsets = cpl_vector_get_data(offsets);

    /* Create the output image list */
    abba = cpl_imagelist_new();
    *abba_off = cpl_vector_new(ngroups);
    pabba_off = cpl_vector_get_data(*abba_off);
    
    /* Loop on the groups */
    for (i=0; i<ngroups; i++) {
        /* Initialise */
        saa = 0;
        /* Create the group list of images */
        group_list = cpl_imagelist_new();
        k = 0;
        for (j=0; j<nima; j++) {
            if (i+1 == groups[j]) {
                /* Get the first offset of the group in abba_off */
                if (k==0) pabba_off[i] = poffsets[j];
                /* To know if we need the saa (shift and add) */
                if (fabs(pabba_off[i]-poffsets[j]) > 1e-3) saa = 1;
                /* Copy the images of the group in group_list */
                tmp_ima = cpl_image_duplicate(cpl_imagelist_get(ilist, j));
                cpl_imagelist_set(group_list, tmp_ima, k);
                tmp_ima = NULL;
                k++;
            }
        }

        if (saa) {
            /* Get the offsets of the group in group_off */
            group_off = cpl_bivector_new(k);
            cpl_vector_fill(cpl_bivector_get_y(group_off), 0.0);
            pgroup_off = cpl_bivector_get_x_data(group_off);
            k = 0;
            for (j=0; j<nima; j++) {
                if (i+1 == groups[j]) {
                    pgroup_off[k] = poffsets[j];
                    k++;
                }
            }
            cpl_vector_subtract_scalar(cpl_bivector_get_x(group_off), 
                    pabba_off[i]);
            /* Shift and add */
            cpl_msg_debug(cpl_func, "Apply shift-and-add for group %d", i+1);
            if ((combined = cpl_geom_img_offset_saa(group_list,
                           group_off, CPL_KERNEL_DEFAULT, 0, 0,
			   CPL_GEOM_FIRST, NULL, NULL)) == NULL) {
                cpl_msg_error(cpl_func,"Cannot shift and add group nb %d", i+1);
                cpl_imagelist_delete(group_list);
                cpl_bivector_delete(group_off);
                cpl_imagelist_delete(abba);
                cpl_vector_delete(*abba_off);
                return NULL;
            }
            cpl_bivector_delete(group_off);
            cpl_image_delete(combined[1]);
            cpl_imagelist_set(abba, combined[0], i);
            cpl_free(combined);
        } else {
            /* Averaging */
            cpl_msg_debug(cpl_func, "Apply averaging for group %d", i+1);
            if ((tmp_ima = cpl_imagelist_collapse_create(group_list)) == NULL) {
                cpl_msg_error(cpl_func, "Cannot average group nb %d", i+1);
                cpl_imagelist_delete(group_list);
                cpl_imagelist_delete(abba);
                cpl_vector_delete(*abba_off);
                return NULL;
            }
            cpl_imagelist_set(abba, tmp_ima, i);
        }
        cpl_imagelist_delete(group_list);
    }
    return abba;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The wavelength calibration
  @param    arc     the arc file 
  @param    oh      the oh lines catalog
  @param    ima     the image with the sky lines 
  @param    raw     the frameset for the header
  @return   int 0 if ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int sofi_spc_jitter_wavecal(
        const char      *   arc,
        const char      *   oh,
        cpl_image       *   ima,
        cpl_frameset    *   raw)
{
    cpl_table           *   arc_tab;
    cpl_polynomial      *   phdisprel;
    cpl_frame           *   cur_frame;
    const char          *   cur_fname;
    cpl_polynomial      *   disprel;
    double                  slit_width;
    cpl_size                p[5];
    double                  xc, a, b, c, d, e;
    int                     degree;
    
    /* Get the wavelength from the arc file */
    if (sofi_spc_jitter_config.wavecal_in == 2) {
        if (arc == NULL) {
            cpl_msg_error(cpl_func, "Missing arc for the wavelength calib");
            return CPL_ERROR_UNSPECIFIED;
        }
        cpl_msg_info(cpl_func, "Get the wavelength from the ARC file");
        if ((arc_tab = cpl_table_load(arc, 1, 0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the arc table");
            sofi_spc_jitter_config.wavecal_out = -1;
            return CPL_ERROR_UNSPECIFIED;
        }
        sofi_spc_jitter_config.wavecal_a0 =
            cpl_table_get_double(arc_tab, "WL_coefficients", 0, NULL);
        sofi_spc_jitter_config.wavecal_a1 = 
            cpl_table_get_double(arc_tab, "WL_coefficients", 1, NULL);
        sofi_spc_jitter_config.wavecal_a2 = 
            cpl_table_get_double(arc_tab, "WL_coefficients", 2, NULL);
        sofi_spc_jitter_config.wavecal_a3 = 
            cpl_table_get_double(arc_tab, "WL_coefficients", 3, NULL);
        sofi_spc_jitter_config.wavecal_a4 = 
            cpl_table_get_double(arc_tab, "WL_coefficients", 4, NULL);
        cpl_table_delete(arc_tab);
        sofi_spc_jitter_config.wavecal_out = 2;
        sofi_spc_jitter_config.wavecal_cc = -1.0;
        return 0;
    }

    /* Get the reference frame */
    cur_frame = cpl_frameset_get_position(raw, 0);
    cur_fname = cpl_frame_get_filename(cur_frame);
    
    /* Get the physical model */
    cpl_msg_info(cpl_func, "Compute the physical model");
    cpl_msg_indent_more();
    if ((phdisprel = sofi_get_disprel_estimate(cur_fname, 1)) == NULL) {
        cpl_msg_error(cpl_func, "cannot compute the physical model");
        sofi_spc_jitter_config.wavecal_out = -1;
        cpl_msg_indent_less();
        return CPL_ERROR_UNSPECIFIED;
    }
    p[0] = 0; p[1] = 1;
    a = cpl_polynomial_get_coeff(phdisprel, p);
    b = cpl_polynomial_get_coeff(phdisprel, p + 1);
    cpl_msg_info(cpl_func, "f(x)=%g + %g*x", a, b);
    sofi_spc_jitter_config.wavecal_a0 = a;
    sofi_spc_jitter_config.wavecal_a1 = b;
    sofi_spc_jitter_config.wavecal_a2 = 0.0;
    sofi_spc_jitter_config.wavecal_a3 = 0.0;
    sofi_spc_jitter_config.wavecal_a4 = 0.0;
    sofi_spc_jitter_config.wavecal_cc = -1.0;
    sofi_spc_jitter_config.wavecal_out = 0;
    cpl_msg_indent_less();

    /* Compute the wavelength using the sky lines */
    if (sofi_spc_jitter_config.wavecal_in == 1) {
        /* Compute the slit_width */
        if ((slit_width = sofi_get_slitwidth(cur_fname)) == -1) {
            cpl_msg_warning(cpl_func, "cannot get the slit width");
            cpl_polynomial_delete(phdisprel);
            return 0;
        }
        /* Compute the wavelength */
        cpl_msg_info(cpl_func, "Compute the wavelength with the sky lines");
        cpl_msg_indent_more();
        if ((disprel = sofi_wavelength_engine(ima, "oh", oh, NULL, NULL, 
                        phdisprel, slit_width, 
                        sofi_spc_jitter_config.wavecal_degree, 
                        sofi_spc_jitter_config.wavecal_err, 
                        sofi_spc_jitter_config.wavecal_nsamples, 
                        sofi_spc_jitter_config.wavecal_ppm, 
                        sofi_spc_jitter_config.display, &xc)) == NULL) {
            cpl_msg_error(cpl_func, "cannot compute the dispersion relation");
            cpl_polynomial_delete(phdisprel);
            cpl_msg_indent_less();
            return 0;
        }
        cpl_msg_info(cpl_func, "Cross correlation factor: %g", xc);
        p[0] = 0; p[1] = 1; p[2] = 2; p[3] = 3; p[4] = 4;
        a = b = c = d = e = 0.0;
        degree = cpl_polynomial_get_degree(disprel);
        a = cpl_polynomial_get_coeff(disprel, p);
        if (degree > 0) b = cpl_polynomial_get_coeff(disprel, p + 1);
        if (degree > 1) c = cpl_polynomial_get_coeff(disprel, p + 2);
        if (degree > 2) d = cpl_polynomial_get_coeff(disprel, p + 3);
        if (degree > 3) e = cpl_polynomial_get_coeff(disprel, p + 4);
        if (degree == 0)
            cpl_msg_info(cpl_func, "f(x)=%g", a);
        if (degree == 1)
            cpl_msg_info(cpl_func, "f(x)=%g + %g*x", a, b);
        if (degree == 2)
            cpl_msg_info(cpl_func, "f(x)=%g + %g*x + %g*x^2", a, b, c);
        if (degree == 3)
            cpl_msg_info(cpl_func, "f(x)=%g + %g*x + %g*x^2 + %g*x^3", 
                    a, b, c, d);
        if (degree == 4)
            cpl_msg_info(cpl_func, "f(x)=%g + %g*x + %g*x^2 + %g*x^3 + %g*x^4", 
                    a, b, c, d, e);
        sofi_spc_jitter_config.wavecal_a0 = a;
        sofi_spc_jitter_config.wavecal_a1 = b;
        sofi_spc_jitter_config.wavecal_a2 = c;
        sofi_spc_jitter_config.wavecal_a3 = d;
        sofi_spc_jitter_config.wavecal_a4 = e;
        sofi_spc_jitter_config.wavecal_cc = xc;
        sofi_spc_jitter_config.wavecal_out = 1;
        cpl_polynomial_delete(disprel);
        cpl_msg_indent_less();
    }
    cpl_polynomial_delete(phdisprel);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create the nodded images
  @param    abba        the abba images list
  @param    abba_off    the abba offsets
  @param    nodded_off  the offsets of the output image list
  @return   the nodded images

  The frames are reduced by pairs. 
  At this stage, we have an even number of frames.

  Each pair (frame n, frame n+1) will generate two nodded frames 
  (frame n-frame n+1) and (frame n+1-frame n).
  Example 
  abba1 - cat A  ---> abba1 - abba2 = nodded1
  abba2 - cat B  ---> abba2 - abba1 = nodded2
  abba3 - cat B  ---> abba3 - abba4 = nodded3
  abba4 - cat A  ---> abba4 - abba3 = nodded4
  abba5 - cat A  ---> abba5 - abba6 = nodded5
  abba6 - cat B  ---> abba6 - abba5 = nodded6
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * sofi_spc_jitter_nodded(
        cpl_imagelist   *   abba,
        cpl_vector      *   abba_off,
        cpl_vector      **  nodded_off)
{
    cpl_imagelist       *   nodded;
    cpl_image           *   tmp_ima;
    int                     nima;
    double                  ref_off;
    int                     i;

    /* Test entries */
    if ((abba == NULL) || (abba_off == NULL)) return NULL;

    /* Initialise */
    nima = cpl_imagelist_get_size(abba);
    if (nima % 2) {
        cpl_msg_error(cpl_func, "Number of images should be even");
        return NULL;
    }
   
    /* Create the offsets between the nodded images */
    *nodded_off = cpl_vector_duplicate(abba_off);
    /* The image list to contain the nodded images */
    nodded = cpl_imagelist_new();
    for (i=0; i<(nima/2); i++) {
        /* a-b */
        tmp_ima = cpl_image_duplicate(cpl_imagelist_get(abba, 2*i));
        cpl_image_subtract(tmp_ima, cpl_imagelist_get(abba, 2*i+1));
        cpl_imagelist_set(nodded, tmp_ima, 2*i);
        /* b-a */
        tmp_ima = cpl_image_duplicate(cpl_imagelist_get(abba, 2*i+1));
        cpl_image_subtract(tmp_ima, cpl_imagelist_get(abba, 2*i));
        cpl_imagelist_set(nodded, tmp_ima, 2*i+1);
    }

    /* Subtract the first offset to the others */
    ref_off = cpl_vector_get(*nodded_off, 0);
    cpl_vector_subtract_scalar(*nodded_off, ref_off);
    return nodded;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Correct the distortion
  @param    ilist       the images list to correct
  @param    arc         the arc file containing the polynomial
  @return   the distortion corrected image list or NULL
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * sofi_spc_jitter_distor(
        cpl_imagelist   *   ilist,
        const char      *   arc)
{
    cpl_polynomial      *   arc_poly;
    cpl_polynomial      *   id_poly;
    cpl_table           *   tab;
    cpl_size                power[2];
    cpl_vector          *   profile;
    cpl_imagelist       *   warped_list;
    cpl_image           *   warped;
    int                     i;
    
    /* Test entries */
    if (ilist == NULL) return NULL;
    if (arc == NULL) return NULL;

    /* Get the arc polynomial */
    arc_poly = cpl_polynomial_new(2);
    cpl_msg_info(cpl_func, "Get the arc distortion from the file");
    if ((tab = cpl_table_load(arc, 1, 0)) == NULL) {
        cpl_msg_error(cpl_func, "cannot load the arc table");
        cpl_polynomial_delete(arc_poly);
        return NULL;
    }
    for (i=0; i<cpl_table_get_nrow(tab); i++) {
        power[0] = cpl_table_get_int(tab, "Degree_of_x", i, NULL);
        power[1] = cpl_table_get_int(tab, "Degree_of_y", i, NULL);
        cpl_polynomial_set_coeff(arc_poly, power,
                cpl_table_get_double(tab, "poly2d_coef", i, NULL));
    }
    cpl_table_delete(tab);

    /* Create the ID polynomial for the other direction */
    id_poly = cpl_polynomial_new(2);
    power[0] = 1;
    power[1] = 0;
    cpl_polynomial_set_coeff(id_poly, power, 1.0);
   
    /* Create the kernel */
    profile = cpl_vector_new(CPL_KERNEL_DEF_SAMPLES);
    cpl_vector_fill_kernel_profile(profile, CPL_KERNEL_DEFAULT,
            CPL_KERNEL_DEF_WIDTH);

    /* Correct the images */
    warped_list = cpl_imagelist_new();
    for (i=0; i<cpl_imagelist_get_size(ilist); i++) {
        warped = cpl_image_duplicate(cpl_imagelist_get(ilist, i));
        if (cpl_image_warp_polynomial(warped, cpl_imagelist_get(ilist, i), 
                    id_poly, arc_poly, profile, CPL_KERNEL_DEF_WIDTH, profile,
                    CPL_KERNEL_DEF_WIDTH) != CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "cannot correct the distortion");
            cpl_image_delete(warped);
            cpl_polynomial_delete(arc_poly);
            cpl_polynomial_delete(id_poly);
            cpl_vector_delete(profile);
            return NULL;
        }
        cpl_imagelist_set(warped_list, warped, i);
    }
    cpl_vector_delete(profile);
    cpl_polynomial_delete(arc_poly);
    cpl_polynomial_delete(id_poly);
    return warped_list;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Refine the offsets
  @param    ima1    the first image
  @param    ima2    the second image
  @return   the computed offset or 10000.0 in error case
 */
/*----------------------------------------------------------------------------*/
static double sofi_spc_jitter_refine_offset(
        cpl_image   *   ima1,
        cpl_image   *   ima2)
{
    double                  pos1, pos2;

    /* Test entries */
    if (ima1 == NULL) return 10000.0;
    if (ima2 == NULL) return 10000.0;
   
    /* Detect the spectra */
    if (irplib_spectrum_find_brightest(ima1, 0.0, NO_SHADOW, 0.0, 1, 
                &pos1) == -1){
        return 10000.0;
    }
    if (irplib_spectrum_find_brightest(ima2, 0.0, NO_SHADOW, 0.0, 1,
                &pos2) == -1){
        return 10000.0;
    }
    return pos1-pos2;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The spectrum extraction
  @param    combined    the combined image
  @return   the extracted table or NULL on error
 */
/*----------------------------------------------------------------------------*/
static cpl_table * sofi_spc_jitter_extract(cpl_image * combined)
{
    cpl_errorstate          allstate = cpl_errorstate_get();
    int                     ri_dist, le_dist, ri_width, le_width, spec_pos;
    int                     nx, ny;
    double                  pos;
    int                     right_side, left_side;
    int                     sky_pos[4];
    cpl_vector          *   sky;
    cpl_vector          *   spec;
    cpl_vector          *   wl;
    double              *   pspec;
    double              *   psky;
    double              *   pwl;
    cpl_table           *   out;
    cpl_bivector        *   toplot;
    int                     throw;
    int                     res;
    int                     i;
    int                     nokfirst = -1;
    int                     noklast = 0;

    /* Test entries */
    if (combined == NULL) return NULL;

    /* Initialise */
    nx = cpl_image_get_size_x(combined);
    ny = cpl_image_get_size_y(combined);
    ri_dist = sofi_spc_jitter_config.extr_sky_ri_dist;
    le_dist = sofi_spc_jitter_config.extr_sky_le_dist;
    ri_width = sofi_spc_jitter_config.extr_sky_ri_width;
    le_width = sofi_spc_jitter_config.extr_sky_le_width;
    spec_pos = sofi_spc_jitter_config.extr_spec_pos;
    res = -1;

    /* Detect the spectrum position if not passed */
    if (spec_pos < 0) {
        if (sofi_spc_jitter_config.throws == NULL) {
            cpl_msg_error(cpl_func, 
                    "Need a throw value to detect the spectra !!");
            return NULL;
        }
        for (i=0; i<cpl_vector_get_size(sofi_spc_jitter_config.throws); i++){
            throw = (int)cpl_vector_get(sofi_spc_jitter_config.throws, i);
            if ((res = irplib_spectrum_find_brightest(combined, throw, 
                            TWO_SHADOWS, 0.0, 1, &pos)) == 0) break;
            if ((res = irplib_spectrum_find_brightest(combined, throw, 
                            ONE_SHADOW, 0.0, 1, &pos)) == 0) break;
        }
        if (res != 0) {
            cpl_msg_error(cpl_func, "Cannot detect the spectrum");
            (void)cpl_error_set(cpl_func, CPL_ERROR_DATA_NOT_FOUND);
            return NULL;
        }
        spec_pos = (int)pos;
        cpl_msg_info(cpl_func, "Spectrum detected at x = %d", spec_pos);
    }

    /* Set the parameters for the extraction */

    /* Spectrum position */
    left_side = spec_pos - (int)(sofi_spc_jitter_config.extr_spec_width/2);
    right_side = left_side + sofi_spc_jitter_config.extr_spec_width;
    if ((left_side < 1) || (right_side > ny)) {
        cpl_msg_error(cpl_func, "Spectrum zone falls outside the image");
        (void)cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                    "%d < 1 || %d > %d", (int)left_side,
                                    (int)right_side, (int)ny);
        return NULL;
    }
    /* Residual Sky position */
    if (ri_dist < 0) ri_dist = 2*sofi_spc_jitter_config.extr_spec_width;
    if (le_dist < 0) le_dist = 2*sofi_spc_jitter_config.extr_spec_width;
    sky_pos[1] = spec_pos - le_dist;
    sky_pos[0] = sky_pos[1] - le_width;
    sky_pos[2] = spec_pos + ri_dist;
    sky_pos[3] = sky_pos[2] + ri_width;

    /* Get the sky */
    sky = cpl_vector_new(ny);
    psky = cpl_vector_get_data(sky);
    if (((sky_pos[0] < 1) || (le_width == 0)) && 
            ((sky_pos[3] <= nx) && (ri_width > 0))) {
        for (i=0; i<ny; i++) {
            psky[i] = cpl_image_get_median_window(combined, sky_pos[2],
                    ny-i, sky_pos[3], ny-i);
        }
    } else if (((sky_pos[3] > nx) || (ri_width == 0))
            && ((sky_pos[0] > 0) && (le_width > 0))) {
        for (i=0; i<ny; i++) {
            psky[i] = cpl_image_get_median_window(combined, sky_pos[0],
                    ny-i, sky_pos[1], ny-i);
        }
    } else if ((ri_width != 0) && (le_width != 0)
            && (sky_pos[0] > 0) && (sky_pos[3] <= nx)) {
        for (i=0; i<ny; i++) {
            psky[i] = cpl_image_get_median_window(combined, sky_pos[2],
                    ny-i, sky_pos[3], ny-i);
            psky[i] += cpl_image_get_median_window(combined, sky_pos[0],
                    ny-i, sky_pos[1], ny-i);
            psky[i] /= 2.0;
        }
    } else {
        for (i=0; i<ny; i++) psky[i] = 0.0;
    }

    /* Estimate the spectrum */
    spec = cpl_vector_new(ny);
    pspec = cpl_vector_get_data(spec);
    for (i=0; i<ny; i++) {
        cpl_errorstate prestate = cpl_errorstate_get();
        pspec[i] = cpl_image_get_flux_window(combined, left_side, ny-i,
                right_side, ny-i);
        if (cpl_errorstate_is_equal(prestate)) {
            if (nokfirst < 0) nokfirst = i;
            noklast = i;
            pspec[i] -= psky[i] * sofi_spc_jitter_config.extr_spec_width;
        } else {
            pspec[i] = 0.0;
            (void)cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                        "%d <= %d. i=%d < %d=ny", (int)left_side,
                                        (int)right_side, (int)i, (int)ny);
        }
    }
    if (nokfirst <= noklast) {
        cpl_msg_warning(cpl_func, "Only %d of %d column(s) processed, from %d "
                        "to %d", (int)(1+noklast-nokfirst), (int)ny,
                        (int)(ny-noklast), (int)(ny-nokfirst));
        cpl_errorstate_dump(allstate, CPL_FALSE,
                            cpl_errorstate_dump_one_warning);
        cpl_errorstate_set(allstate);
    } else {
        (void)cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                    "All %d columns failed", (int)ny);
    }

    /* Get the wavelength */
    wl = cpl_vector_new(ny);
    pwl = cpl_vector_get_data(wl);
    for (i=0; i<ny; i++) {
        pwl[i] = sofi_spc_jitter_config.wavecal_a0 +
            sofi_spc_jitter_config.wavecal_a1 * (i+1) +
            sofi_spc_jitter_config.wavecal_a2 * (i+1) * (i+1) +
            sofi_spc_jitter_config.wavecal_a3 * (i+1) * (i+1) * (i+1) +
            sofi_spc_jitter_config.wavecal_a4 * (i+1) * (i+1) * (i+1) + (i+1);
    }

    /* Plot the spectrum if requested */
    if (sofi_spc_jitter_config.display) {
        toplot = cpl_bivector_wrap_vectors(wl, spec);
        cpl_plot_bivector(NULL, "t 'Spectrum' w lines", NULL, toplot);
        cpl_bivector_unwrap_vectors(toplot);
        toplot = cpl_bivector_wrap_vectors(wl, sky);
        cpl_plot_bivector(NULL, "t 'Sky' w lines", NULL, toplot);
        cpl_bivector_unwrap_vectors(toplot);
    }
    
    /* Create and fill the output table */
    out = cpl_table_new(nx);
    cpl_table_new_column(out, "Y_coordinate", CPL_TYPE_DOUBLE);
    cpl_table_new_column(out, "Extracted_spectrum_value", CPL_TYPE_DOUBLE);
    cpl_table_new_column(out, "Sky_spectrum", CPL_TYPE_DOUBLE);
    for (i=0; i<nx; i++) {
        cpl_table_set_double(out, "Y_coordinate", i, pwl[i]);
        cpl_table_set_double(out, "Extracted_spectrum_value", i, pspec[i]);
        cpl_table_set_double(out, "Sky_spectrum", i, psky[i]);
    }
    cpl_vector_delete(wl);
    cpl_vector_delete(spec);
    cpl_vector_delete(sky);
    return out;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the recipe products on disk
  @param    set     the input frame set
  @param    ima     image to save
  @param    tab     table to save
  @param    parlist the input list of parameters
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code sofi_spc_jitter_save(cpl_frameset            * set,
                                           const cpl_image         * ima,
                                           const cpl_table         * tab,
                                           const cpl_parameterlist * parlist)
{
    cpl_propertylist * plist;
    /* Get the QC params in qclist */
    cpl_propertylist * qclist = cpl_propertylist_new();
    cpl_propertylist * paflist;
    /* Get the reference frame */
    const cpl_frame  * ref_frame
        = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);
    char                    qc_str[128];
    int                     i;


    /* Add QC parameters */
    if (sofi_spc_jitter_config.filter[0] != (char)0)
        cpl_propertylist_append_string(qclist, "ESO QC FILTER OBS",
                sofi_spc_jitter_config.filter);
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO1",
            sofi_spc_jitter_config.wavecal_a0);
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO2",
            sofi_spc_jitter_config.wavecal_a1);
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO3",
            sofi_spc_jitter_config.wavecal_a2);
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO4",
            sofi_spc_jitter_config.wavecal_a3);
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO5",
            sofi_spc_jitter_config.wavecal_a4);
    cpl_propertylist_append_double(qclist, "ESO QC WLEN",
            sofi_spc_jitter_config.wavecal_a0 +
            sofi_spc_jitter_config.wavecal_a1 * 512 +
            sofi_spc_jitter_config.wavecal_a2 * 512 * 512 +
            sofi_spc_jitter_config.wavecal_a3 * 512 * 512 * 512 +
            sofi_spc_jitter_config.wavecal_a4 * 512 * 512 * 512 * 512);
    cpl_propertylist_append_double(qclist, "ESO QC DISP XCORR",
            sofi_spc_jitter_config.wavecal_cc);
    if (sofi_spc_jitter_config.wavecal_out == 0) {
        cpl_propertylist_append_string(qclist, "ESO QC WLMETHOD",
                "physical model");
    } else if (sofi_spc_jitter_config.wavecal_out == 1) {
        cpl_propertylist_append_string(qclist, "ESO QC WLMETHOD",
                "sky lines");
    } else if (sofi_spc_jitter_config.wavecal_out == 2) {
        cpl_propertylist_append_string(qclist, "ESO QC WLMETHOD",
                "arc file");
    }
    for (i=0; i<cpl_vector_get_size(sofi_spc_jitter_config.intensities);i++) {
        sprintf(qc_str, "ESO QC SPEC INTENS%d", i+1);
        cpl_propertylist_append_double(qclist, qc_str,
                cpl_vector_get(sofi_spc_jitter_config.intensities, i));
    }

    /* Change WCS keywords to the computed wavelength solution */
    cpl_propertylist_update_double(qclist, "CRVAL1",
            sofi_spc_jitter_config.wavecal_a0);
    cpl_propertylist_update_double(qclist, "CRVAL2", 1.0);
    cpl_propertylist_update_double(qclist, "CRPIX1", 1.0);
    cpl_propertylist_update_double(qclist, "CRPIX2", 1.0);
    cpl_propertylist_update_double(qclist, "CDELT1",
            sofi_spc_jitter_config.wavecal_a1);
    cpl_propertylist_update_double(qclist, "CDELT2", 1.0);
    cpl_propertylist_update_string(qclist, "CTYPE1", "LINEAR");
    cpl_propertylist_update_string(qclist, "CTYPE2", "LINEAR");
    if (cpl_propertylist_has(qclist, "CD1_1")) {
        cpl_propertylist_update_double(qclist, "CD1_1",
                sofi_spc_jitter_config.wavecal_a1);
    } else {
        cpl_propertylist_insert_after_double(qclist, "CTYPE2", "CD1_1",
                sofi_spc_jitter_config.wavecal_a1);
    }
    if (cpl_propertylist_has(qclist, "CD2_2")) {
        cpl_propertylist_update_double(qclist, "CD2_2", 1.0);
    } else {
        cpl_propertylist_insert_after_double(qclist, "CD1_1", "CD2_2", 1.0);
    }

    /* Write the image  */
    irplib_dfs_save_image(set, parlist, set, ima, CPL_BPP_IEEE_FLOAT,
                          RECIPE_STRING, SOFI_SPC_JITTER_COMB, qclist,
                          NULL, PACKAGE "/" PACKAGE_VERSION,
                          RECIPE_STRING "_combined" CPL_DFS_FITS);

    /* Write the table  */
    if (tab != NULL) {
        irplib_dfs_save_table(set, parlist, set, tab, NULL, RECIPE_STRING,
                SOFI_SPC_JITTER_EXTR, qclist, NULL, PACKAGE "/" PACKAGE_VERSION,
                RECIPE_STRING "_extracted" CPL_DFS_FITS);
    }


    /* Get FITS header from reference file */
    plist = cpl_propertylist_load(cpl_frame_get_filename(ref_frame), 0);
    if (plist == NULL) {
        cpl_propertylist_delete(qclist);
        return cpl_error_set_message(cpl_func, cpl_error_get_code(), "Could "
                                     "not read header from reference frame");
    }

    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist,
        "^(ARCFILE|MJD-OBS|INSTRUME|ESO TPL ID|ESO TPL NEXP|ESO DPR CATG|"
        "ESO DPR TECH|ESO DPR TYPE|DATE-OBS|ESO INS GRAT NAME|"
        "ESO INS GRAT WLEN|ESO INS OPTI1 ID|ESO OBS ID|ESO OBS TARG NAME)$", 0);
    cpl_propertylist_delete(plist);

    /* Copy the QC in paflist */
    cpl_propertylist_copy_property_regexp(paflist, qclist, ".", 0);
    cpl_propertylist_delete(qclist);

    /* Save the PAF file */
    cpl_dfs_save_paf("SOFI", RECIPE_STRING, paflist,
                     RECIPE_STRING CPL_DFS_PAF);
    cpl_propertylist_delete(paflist);

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    compares two doubles in regard to a given threshold
  @param    off1 first double
  @param    off2 second double
  @param    thresh  threshold
  @return   1 if threshold is between offsets, 0 otherwise
 */
/*----------------------------------------------------------------------------*/
static int off_comp(double off1, double off2, double thresh)
{
    if (((off1>thresh) && (off2<thresh)) || ((off1<thresh) && (off2>thresh)))
        return 1;
    else return 0;
}

